# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to use command line flags with Fiddle Buildables.

While it's generally better to check in the full configuration into a source-
control system, there are some instances (e.g. hyperparameter sweeps) where it's
most effective to set initialization parameters (hyperparameters) from command
line flags. This library exposes functions you can call within your program.


## Quickstart

> Note: although different bits of functionality are designed to be usable
> piecemeal, this is the fastest way to get all flag-integrated features.


First, some definitions:

 - *A module for base configurations*: So that libraries can be easily reused
   with or without fiddle, code to define fiddle configurations are often put
   in a separate package / module. Most commonly, there is one function (called
   `default` by convention)
 - *Fiddlers*: A fiddler is a function that takes a `fdl.Buildable` and performs
   arbitrary mutations. These are commonly also grouped into the same module
   that defines the base configuration(s).
 - *Overrides*: Overrides are command line options that set values to
   parameters within the (nested) fiddle tree.

The simplest way to use the fiddle support for absl flags is to pass
`flags_parser=absl_flags.flags_parser` when calling `app.run`. Then, simply call
`cfg = absl_flags.create_buildable_from_flags(configs)` where `configs` is a
module for base configurations and fiddlers.

For example:


```py
## ---- config.py ----
import fiddle as fdl

def baseline() -> fdl.Config:  # A base configuration.
  my_model = fdl.Config(MyModel)
  my_optimizer = fdl.Config(MyOptimizer)
  return fdl.Config(MyTrainer, optimizer=my_optimizer, model=my_model)

def set_dtypes_to_bfloat_16(cfg: fdl.Config):  # A fiddler.
  cfg.model.dtype = 'bfloat16'  # Arbitrary python.

##  ---- main.py ----
from absl import app
import fiddle as fdl
from fiddle import absl_flags
import config

def main(argv):
  cfg = absl_flags.create_buildable_from_flags(configs)
  trainer = fdl.build(cfg)
  trainer.train()

if __name__ == '__main__':
  app.run(main, flags_parser=absl_flags.flags_parser)
```

## Overrides-only

If you simply want to use command line flags to override values in a config
built through arbitrary python, follow these steps:

 1. *main*: Pass `fdl.absl_flags.flags_parser` when calling `absl.app.run`. For
    example, your `app.run` call should look like the following:
        `app.run(my_main_function, flags_parser=fiddle.absl_flags.flags_parser)`
 2. *Augment a config*: Somewhere in your program, call
    `fdl.absl_flags.apply_overrides_to(cfg)` to apply settings defined on the
    command line to `cfg`. (This is often best done right before the call to
    `fdl.build(cfg)`.)
 3. Set values on the command line: `--fdl.my.property=3 --fdl.other.setting=5`

### Configuring multiple fiddle objects

If you have multiple fiddle objects that you'd like to configure on the command
line, create an empty top-level config and use that. Example:

```
my_program_part_1 = fdl.Config(...)
my_program_part_2 = fdl.Config(...)

top_level = fiddle.experimental.DictConfig(part1=my_program_part_1)
top_level.part2 = my_program_part_2
fdl.absl_flags.apply_overrides_to(top_level)

run_part_1(fdl.build(my_program_part_1))
run_part_2(fdl.build(my_program_part_2))
```

## Acknowledgements

This implementation has drawn inspiration from multiple sources, including
[T5X](https://github.com/google-research/t5x), among others. Thank you!
"""

import ast
import dataclasses
import inspect
import re
import sys
import typing
from typing import Any, List, Union, Sequence

from absl import app
from absl import flags
from absl import logging

from fiddle import config
from fiddle import module_reflection
from fiddle import printing
from fiddle import tagging
from fiddle.experimental import selectors

_FDL_CONFIG = flags.DEFINE_string(
    'fdl_config',
    default=None,
    help='The name of the no-arity function to construct a fdl.Buildable.')
_FIDDLER = flags.DEFINE_multi_string(
    'fiddler', default=[], help='Fiddlers are functions that tweak a config.')
_FDL_SET = flags.DEFINE_multi_string(
    'fdl_set', default=[], help='Fiddle configuration settings.')
_FDL_TAGS_SET = flags.DEFINE_multi_string(
    'fdl_tags_set',
    default=[],
    help='Fiddle tags setting, by name. Typically accessed via the alias '
    '--fdl_tag.foo.Bar=123')
_FDL_HELP = flags.DEFINE_bool(
    'fdl_help', False, help='Print out the flags-built config and exit.')


@dataclasses.dataclass
class _IndexKey:
  """Wraps a string or integer index into a dictionary or list/tuple."""
  key: Union[str, int]

  def __post_init__(self):
    try:
      self.key = ast.literal_eval(self.key)
    except Exception:  # pylint: disable=broad-except
      pass  # leave self as-is.

  def __call__(self, obj):
    return obj[self.key]

  def update(self, obj, new_value):
    """Updates `obj`'s `self.key` to `new_value`."""
    obj[self.key] = new_value


@dataclasses.dataclass
class _AttributeKey:
  """Wraps the name of an attribute."""
  __slots__ = ('name')
  name: str

  def __call__(self, obj):
    return getattr(obj, self.name)

  def update(self, obj, new_value):
    setattr(obj, self.name, new_value)


_PATH_COMPONENT_REGEX = re.compile(r'(?:\.([A-Za-z_][^\.\[]*))|\[([^]]+)\]')


def _parse_path(path: str) -> List[Union[_AttributeKey, _IndexKey]]:
  """Parses a path into a list of either attributes or index lookups."""
  path = f'.{path}'  # Add a leading `.` to make parsing work properly.
  result = []
  curr_index = 0
  while curr_index < len(path):
    match = _PATH_COMPONENT_REGEX.match(path, curr_index)
    if match is None:
      raise ValueError(
          f'Could not parse {path[1:]} (failed index: {curr_index - 1}).')
    curr_index = match.end(0)  # Advance
    if match[1] is not None:
      result.append(_AttributeKey(match[1]))
    else:
      result.append(_IndexKey(match[2]))
  return result


def apply_overrides_to(cfg: config.Buildable):
  """Applies all command line flags to `cfg`."""
  for flag in _FDL_SET.value:
    path, value = flag.split('=', maxsplit=1)
    *parents, last = _parse_path(path)
    walk = typing.cast(Any, cfg)
    try:
      for parent in parents:
        walk = parent(walk)
    except Exception:  # pylint: disable=broad-except
      raise ValueError(f'Invalid path "{path}".')  # pylint: disable=raise-missing-from
    try:
      last.update(walk, ast.literal_eval(value))
    except:
      raise ValueError(f'Could not set "{value}" path "{path}".')  # pylint: disable=raise-missing-from


def _rewrite_fdl_args(args: Sequence[str]) -> List[str]:
  """Rewrites short-form Fiddle flags.

  There are two main rewrites:

    * `--fdl.NAME=VALUE` to `--fdl_set=NAME = VALUE`.
    * `--fdl_tag.NAME=VALUE` to `--fdl_tags_set=NAME = VALUE`.

  Args:
    args: Command-line args.

  Returns:
    Rewritten args.
  """

  def _rewrite(arg: str) -> str:
    if arg.startswith('--fdl.') or arg.startswith('--fdl_tag.'):
      if '=' not in arg:
        prefix = arg.split('.', maxsplit=1)[0]
        raise ValueError(
            f'Fiddle setting must be of the form `{prefix}.NAME=VALUE`; '
            f'`got: "{arg}".')
      if arg.startswith('--fdl.'):
        explicit_name = 'fdl_set'
      elif arg.startswith('--fdl_tag.'):
        explicit_name = 'fdl_tags_set'
      _, arg = arg.split('.', maxsplit=1)  # Strip --fdl. or --fdl_tag. prefix.
      path, value = arg.split('=', maxsplit=1)
      rewritten = f'--{explicit_name}={path}={value}'
      logging.debug('Rewrote flag "%s" to "%s".', arg, rewritten)
      return rewritten
    else:
      return arg

  return list(map(_rewrite, args))


def flags_parser(args: Sequence[str]):
  """Flag parser.

  See absl.app.parse_flags_with_usage and absl.app.main(..., flags_parser).

  Args:
    args: All command line arguments.

  Returns:
    Whatever `absl.app.parse_flags_with_usage` returns. Sorry!
  """
  return app.parse_flags_with_usage(_rewrite_fdl_args(args))


def set_tags(cfg: config.Buildable):
  """Sets tags based on their name, from CLI flags."""
  all_tags = tagging.list_tags(cfg, add_superclasses=True)
  for flag in _FDL_TAGS_SET.value:
    name, value = flag.split('=', maxsplit=1)
    matching_tags = [tag for tag in all_tags if tag.name == name]
    if not matching_tags:
      # TODO: Improve and unify these errors.
      loose_matches = [
          tag for tag in all_tags
          if name.lower().replace('_', '') in tag.name.lower().replace('_', '')
      ]
      did_you_mean = f' Did you mean {loose_matches}?' if loose_matches else ''
      raise ValueError(f'No tags with name {name!r} in config.{did_you_mean}')
    elif len(matching_tags) > 1:
      raise EnvironmentError(
          f'There were multiple tags with name {name}; perhaps some module '
          'reloading weirdness is going on? This should be very rare. Please '
          'make sure that you are not dynamically creating subclasses of '
          '`fdl.Tag` and using 2 instances with the same module and name '
          'in the configuration.')
    selectors.select(
        cfg, tag=matching_tags[0]).set(value=ast.literal_eval(value))


def apply_fiddlers_to(cfg: config.Buildable, source_module: Any):
  """Applies fiddlers to `cfg`."""
  # TODO: Consider allowing arbitrary imports using importlib.
  for fiddler_name in _FIDDLER.value:
    if not hasattr(source_module, fiddler_name):
      available_fiddlers = ', '.join(
          module_reflection.find_fiddler_like_things(source_module))
      raise ValueError(
          f'No fiddler named {fiddler_name} found; available fiddlers: '
          f'{available_fiddlers}.')
    fiddler = getattr(source_module, fiddler_name)
    fiddler(cfg)


def create_buildable_from_flags(module: Any) -> config.Buildable:
  """Returns a fdl.Buildable based on standardized flags.

  Args:
    module: A common namespace to use as the basis for finding configs and
      fiddlers.
  """
  # TODO: Explore allowing arbitrary imports.
  base_name = _FDL_CONFIG.value
  if _FDL_HELP.value or base_name is None:
    print(
        'Available symbols (may be base configs or fiddlers):\n',
        file=sys.stderr)
    for key in dir(module):
      # Ignore private and module objects.
      if key.startswith('_'):
        continue
      obj = getattr(module, key)
      if inspect.ismodule(obj):
        continue

      if hasattr(obj, '__doc__') and obj.__doc__:
        obj_doc = '\n'.join(obj.__doc__.strip().split('\n\n')[:2])
        obj_doc = re.sub(r'\s+', ' ', obj_doc)
        docstring = f' - {obj_doc}'
      else:
        docstring = ''
      print(f'  {key}{docstring}', file=sys.stderr)
  print(file=sys.stderr)
  if base_name is None:
    raise app.UsageError('--fdl_config is required.')
  if not hasattr(module, base_name):
    available_names = module_reflection.find_base_config_like_things(module)
    raise ValueError(f'Could not init a buildable from {base_name}; '
                     f'available names: {", ".join(available_names)}.')
  base_fn = getattr(module, base_name)
  if hasattr(base_fn, 'as_buildable'):
    buildable = base_fn.as_buildable()
  else:
    buildable = base_fn()
  apply_fiddlers_to(buildable, source_module=module)
  set_tags(buildable)
  apply_overrides_to(buildable)
  if _FDL_HELP.value:
    print('Tags (override as --fdl_tag.<name>=<value>):\n\n', file=sys.stderr)
    for tag in tagging.list_tags(buildable):
      print(f' - {tag.name} ({tag.description})', file=sys.stderr)
    print(
        'Config values (override as --fdl.<name>=<value>):\n\n ',
        printing.as_str_flattened(buildable).replace('\n', '\n  '),
        file=sys.stderr)
    sys.exit()
  return buildable



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library for manipulating selections of a Buildable DAG.

A common need for configuration libraries is to override settings in some kind
of base configuration, and these APIs allow such overrides to take place
imperatively.
"""

from typing import Any, Callable, Iterator, Optional, Set, Type, Union

from fiddle import config
from fiddle import tagging
import tree

# Maybe DRY up with type declaration in autobuilders.py?
FnOrClass = Union[Callable[..., Any], Type[Any]]


class Selection:
  """Represents a selection of nodes.

  This selection is declarative, so if subtrees / subgraphs of `cfg` change and
  later match or don't match, a different set of nodes will be returned.

  Generally this class is intended for modifying attributes of a buildable DAG
  in a way that doesn't alter its structure. We do not pay particular attention
  to structure-altering modifications right now; please do not depend on such
  behavior.
  """
  cfg: config.Buildable
  fn_or_cls: Optional[FnOrClass]
  tag: Optional[tagging.TagType]
  match_subclasses: bool
  buildable_type: Type[config.Buildable]

  def __init__(
      self,
      cfg: config.Buildable,
      fn_or_cls: Optional[FnOrClass] = None,
      *,
      tag: Optional[tagging.TagType] = None,
      match_subclasses: bool = True,
      buildable_type: Optional[Type[config.Buildable]] = None,
  ):
    super().__setattr__("cfg", cfg)
    super().__setattr__("fn_or_cls", fn_or_cls)
    super().__setattr__("tag", tag)
    super().__setattr__("match_subclasses", match_subclasses)

    if buildable_type is None:
      # Set `buildable_type` for implementation in `_matches` below. In general
      # buildable_type should only be None when `tag` is set and `fn_or_cls` is
      # not.
      if tag is None or fn_or_cls is not None:
        buildable_type = config.Buildable

    super().__setattr__("buildable_type", buildable_type)

  def _matches(self, node: config.Buildable) -> bool:
    """Helper for __iter__ function, determining if a node matches."""

    # Implementation note: To allow for future expansion of this class, checks
    # here should be expressed as `if not my_matcher.match(x): return False`.

    # For tags, check that the tag is present, and then
    # refer to underlying configuration.
    if self.tag is not None:
      if not isinstance(node, tagging.TaggedValue):
        return False
      if not any(issubclass(tag, self.tag) for tag in node.tags):
        return False
      node = node.value

    if self.buildable_type is not None:
      if not isinstance(node, self.buildable_type):
        return False

    if self.fn_or_cls is not None:
      if self.fn_or_cls != node.__fn_or_cls__:
        # Determines if subclass matching is allowed, and if the node is a
        # subclass of `self.fn_or_cls`. We check whether both are instances
        # of `type` to avoid `issubclass` errors when either side is actually a
        # function.
        is_subclass = (
            self.match_subclasses  #
            and isinstance(self.fn_or_cls, type)  #
            and isinstance(node.__fn_or_cls__, type)  #
            and issubclass(node.__fn_or_cls__, self.fn_or_cls))
        if not is_subclass:
          return False

    return True

  def _iter_helper(self, node: config.Buildable, seen: Set[int]):
    """Helper for __iter__ function below, keeping track of seen nodes."""
    if id(node) in seen:
      return
    seen.add(id(node))

    if self._matches(node):
      yield node

    for leaf in tree.flatten(node.__arguments__):
      if isinstance(leaf, config.Buildable):
        yield from self._iter_helper(leaf, seen)

  def __iter__(self) -> Iterator[config.Buildable]:
    """Iterates over nodes in the tree.

    Returns:
      config.Buildable nodes matching this selection.
    """
    return self._iter_helper(self.cfg, set())

  def __setattr__(self, name: str, value: Any) -> None:
    """Shorthand to set a single value.

    Args:
      name: Name of the attribute to set.
      value: Value to set on all matching nodes.
    """
    for matching in self:
      setattr(matching, name, value)

  def set(self, **kwargs) -> None:
    """Sets multiple attributes on nodes matching this selection.

    Args:
      **kwargs: Properties to set on matching nodes.
    """
    for matching in self:
      for name, value in kwargs.items():
        setattr(matching, name, value)

  def get(self, name: str) -> Iterator[Any]:
    """Gets all values for a particular attribute.

    Args:
      name: Name of the attribute on matching nodes.

    Yields:
      Values configured for the attribute with name `name` on matching nodes.
    """
    for matching in self:
      yield getattr(matching, name)


select = Selection  # pylint: disable=invalid-name



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the `Config` class and associated subclasses and functions."""
from __future__ import annotations

import abc
import collections
import copy
import functools
import inspect
from typing import Any, Callable, Collection, Dict, Generic, List, Type, TypeVar, Union

from fiddle import history

T = TypeVar('T')
TypeOrCallableProducingT = Union[Callable[..., T], Type[T]]

# Unique object instance that should never be used by end-users, and can thus
# be used to differentiate between unset values and user-set values that are
# None or other commonly-used sentinel.
_UNSET_SENTINEL = object()


class Buildable(Generic[T], metaclass=abc.ABCMeta):
  """Base class for buildable types (`Config` and `Partial`).

  Buildable types implement a `__build__` method that is called during
  `fdl.build()` with arguments set on the `Buildable` to get output for the
  corresponding instance.
  """

  # "Dunder"-style names were chosen here to reduce the possibility of naming
  # conflicts while still making the fields accessible.
  __fn_or_cls__: TypeOrCallableProducingT
  __signature__: inspect.Signature
  __arguments__: Dict[str, Any]
  __argument_history__: Dict[str, List[history.HistoryEntry]]
  _has_var_keyword: bool

  def __init__(self, fn_or_cls: Union['Buildable', TypeOrCallableProducingT],
               *args, **kwargs):
    """Initialize for `fn_or_cls`, optionally specifying parameters.

    Args:
      fn_or_cls: The function or class to configure, or a `Buildable` to copy.
      *args: Any positional arguments to configure for `fn_or_cls`.
      **kwargs: Any keyword arguments to configure for `fn_or_cls`.
    """
    # Using `super().__setattr__` here because assigning directly would trigger
    # our `__setattr__` override. Using `super().__setattr__` instead of special
    # casing these attribute names in `__setattr__` also has the effect of
    # making them easily gettable but not as easily settable by user code.
    if isinstance(fn_or_cls, Buildable):
      initial_arguments = fn_or_cls.__arguments__.copy()
      fn_or_cls = fn_or_cls.__fn_or_cls__
    else:
      initial_arguments = {}
    super().__setattr__('__fn_or_cls__', fn_or_cls)
    super().__setattr__('__arguments__', initial_arguments)
    signature = inspect.signature(fn_or_cls)
    super().__setattr__('__signature__', signature)
    has_var_keyword = any(param.kind == param.VAR_KEYWORD
                          for param in signature.parameters.values())
    arg_history = collections.defaultdict(list)
    arg_history['__fn_or_cls__'].append(
        history.entry('__fn_or_cls__', fn_or_cls))
    super().__setattr__('__argument_history__', arg_history)
    super().__setattr__('_has_var_keyword', has_var_keyword)

    arguments = signature.bind_partial(*args, **kwargs).arguments
    for name in list(arguments.keys()):  # Make a copy in case we mutate.
      param = signature.parameters[name]
      if param.kind == param.VAR_POSITIONAL:
        # TODO: Add *args support.
        err_msg = 'Variable positional arguments (aka `*args`) not supported.'
        raise NotImplementedError(err_msg)
      elif param.kind == param.VAR_KEYWORD:
        arguments.update(arguments.pop(param.name))

    if hasattr(fn_or_cls, '__fiddle_init__'):
      fn_or_cls.__fiddle_init__(self)

    for name, value in arguments.items():
      setattr(self, name, value)

  @abc.abstractmethod
  def __build__(self, *args, **kwargs):
    """Builds output for this instance; see subclasses for details."""
    raise NotImplementedError()

  def __getattr__(self, name: str):
    """Get parameter with given `name`."""
    value = self.__arguments__.get(name, _UNSET_SENTINEL)
    if value is not _UNSET_SENTINEL:
      return value
    param = self.__signature__.parameters.get(name)
    if param is not None and param.default is not param.empty:
      return param.default
    msg = f"No parameter '{name}' has been set on {self!r}."
    # TODO: Implement an edit distance function and display valid
    # attributes that are close to `name`.
    if hasattr(self.__fn_or_cls__, name):
      msg += (f' Note: {self.__fn_or_cls__.__module__}.'
              f'{self.__fn_or_cls__.__qualname__} has an attribute/method with '
              'this name, so this could be caused by using a '
              f'fdl.{self.__class__.__qualname__} in '
              'place of the actual function or class being configured. Did you '
              'forget to call `fdl.build(...)`?')
    raise AttributeError(msg)

  def __setattr__(self, name: str, value: Any):
    """Sets parameter `name` to `value`."""
    param = self.__signature__.parameters.get(name)

    if param is not None:
      if param.kind == param.POSITIONAL_ONLY:
        # TODO: Add positional-only arg support.
        raise NotImplementedError(
            'Positional only arguments not supported. '
            f'Tried to set {name!r} on {self.__fn_or_cls__}')
      elif param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
        # Just pretend it doesn't correspond to a valid parameter name... below
        # a TypeError will be thrown unless there is a **kwargs parameter.
        param = None

    if param is None and not self._has_var_keyword:
      if name in self.__signature__.parameters:
        err_msg = (f'Variadic arguments (e.g. *{name}) are not supported.')
      else:
        valid_parameter_names = (
            name for name, param in self.__signature__.parameters.items()
            if param.kind not in (param.POSITIONAL_ONLY, param.VAR_POSITIONAL))
        err_msg = (f"No parameter named '{name}' exists for "
                   f'{self.__fn_or_cls__}; valid parameter names: '
                   f"{', '.join(valid_parameter_names)}.")
      raise TypeError(err_msg)

    self.__arguments__[name] = value
    self.__argument_history__[name].append(history.entry(name, value))

  def __delattr__(self, name):
    """Unsets parameter `name`."""
    try:
      del self.__arguments__[name]
      entry = history.entry(name, history.DELETED)
      self.__argument_history__[name].append(entry)
    except KeyError:
      err = AttributeError(f"No parameter '{name}' has been set on {self!r}")
      raise err from None

  def __dir__(self) -> Collection[str]:
    """Provide a useful list of attribute names, optimized for Jupyter/Colab.

    `__dir__` is often implicitly called by tooling such as Jupyter/Colab to
    provide autocomplete suggestions. This implementation of `__dir__` makes it
    easy to see what are valid attributes to get, set, or delete.

    Returns:
      A list of useful attribute names corresponding to set or unset parameters.
    """
    valid_param_names = {
        name for name, param in self.__signature__.parameters.items()
        if param.kind in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY)
    }
    set_argument_names = self.__arguments__.keys()
    all_names = valid_param_names.union(set_argument_names)
    return all_names

  def __repr__(self):
    if hasattr(self.__fn_or_cls__, '__qualname__'):
      formatted_fn_or_cls = self.__fn_or_cls__.__qualname__
    else:
      formatted_fn_or_cls = str(self.__fn_or_cls__)
    formatted_params = [f'{k}={v!r}' for k, v in self.__arguments__.items()]
    name = type(self).__name__
    return f"<{name}[{formatted_fn_or_cls}({', '.join(formatted_params)})]>"

  def __copy__(self):
    """Shallowly copies this `Buildable` instance.

    This copy implementation ensures that setting parameters on a copy of a
    `Buildable` won't affect the original instance. However, the copy is
    shallow, so parameter values will still refer to the same instances of
    objects after the copy.

    Returns:
      A shallow copy of this `Buildable`.
    """
    return type(self)(self)  # pytype: disable=not-instantiable

  def __deepcopy__(self, memo):
    """Deeply copies this `Buildable` instance.

    This copy implementation ensures that setting parameters on a copy of a
    `Buildable` won't affect the original instance.

    Args:
      memo: The `deepcopy` memoization dictionary (used to avoid copying the
        same instance of an object more than once).

    Returns:
      A deep copy of this `Buildable`.
    """
    config_copy = type(self)(self.__fn_or_cls__)  # pytype: disable=not-instantiable
    deepcopied_arguments = copy.deepcopy(self.__arguments__, memo)
    object.__setattr__(config_copy, '__arguments__', deepcopied_arguments)
    return config_copy

  def __eq__(self, other):
    """Returns true iff self and other contain the same argument values."""
    if not isinstance(other, self.__class__):
      return False
    if self.__fn_or_cls__ is not other.__fn_or_cls__:
      return False
    assert self._has_var_keyword == other._has_var_keyword, (
        'Internal invariant violated: has_var_keyword should be the same if '
        "__fn_or_cls__'s are the same.")

    return self.__arguments__ == other.__arguments__

  def __getstate__(self):
    """Gets pickle serialization state, removing some fields.

    For now, we discard the __signature__ (which can be recalculated) and
    __argument_history__, because these tend to contain values which cannot be
    serialized.

    Returns:
      Dict of serialized state.
    """
    result = dict(self.__dict__)
    result.pop('__signature__', None)
    return result

  def __setstate__(self, state) -> None:
    """Loads pickle serialization state.

    This re-derives the signature if not present, and adds an empty
    __argument_history__, if it was removed.

    Args:
      state: State dictionary, from __getstate__.
    """
    self.__dict__.update(state)  # Support unpickle.
    if '__signature__' not in self.__dict__:
      object.__setattr__(self, '__signature__',
                         inspect.signature(self.__fn_or_cls__))


class Config(Generic[T], Buildable[T]):
  """A mutable representation of a function or class's parameters.

  This class represents the configuration for a given function or class,
  exposing configured parameters as mutable attributes. For example, given a
  class

      TestClass:
        '''Example class for demonstration purposes.'''

        def __init__(self, arg, kwarg=None):
          self.arg = arg
          self.kwarg = kwarg

  a configuration may (for instance) be accomplished via

      class_config = Config(TestClass, 1, kwarg='kwarg')

  or via

      class_config = Config(TestClass)
      class_config.arg = 1
      class_config.kwarg = 'kwarg'

  A function can be configured in the same ways:

      def test_function(arg, kwarg=None):
        return arg, kwarg

      fn_config = Config(test_function, 1)
      fn_config.kwarg = 'kwarg'

  A `Config` instance may be transformd into instances and function outputs by
  passing it to the `build` function. The `build` function invokes each function
  or class in the configuration tree (appropriately propagating the built
  outputs from nested `Config`s). For example, using the `TestClass` config from
  above:

      instance = build(class_config)
      assert instance.arg == 'arg'
      assert instance.kwarg == 'kwarg'

  If the same `Config` instance is used in multiple places within the
  configuration tree, its function or class is invoked only once during `build`,
  and the result shared across all occurences of the `Config` instance. (See
  `build` documentation for further details.) To create a new instance of a
  `Config` with the same parameter settings that will yield a separate instance
  during `build`, `copy.copy()` or `copy.deepcopy()` may be used.

  A class or function can customize the Config instance by defining a
  `__fiddle_init__` property. For example:

      class MyClass:
        def __init__(self, x, y, z):
          ...

        @staticmethod
        def __fiddle_init__(cfg):
          cfg.y = 42
          cfg.z = Config(MyOtherClass)
  """

  # NOTE: We currently need to repeat these annotations for pytype.
  __fn_or_cls__: TypeOrCallableProducingT
  __signature__: inspect.Signature

  def __build__(self, *args, **kwargs):
    """Builds this `Config` for the given `args` and `kwargs`.

    This method is called during `build` to get the output for this `Config`.

    Args:
      *args: Positional arguments to pass to `self.__fn_or_cls__`.
      **kwargs: Keyword arguments to pass to `self.__fn_or_cls__`.

    Returns:
      The result of calling `self.__fn_or_cls__` with the given `args` and
      `kwargs`.
    """
    return self.__fn_or_cls__(*args, **kwargs)


class Partial(Generic[T], Buildable[T]):
  """A `Partial` config creates a partial function or class when built.

  In some cases, it may be desired to leave a function or class uninvoked, and
  instead output a corresponding `functools.partial` object. Where the `Config`
  base class calls its underlying `__fn_or_cls__` when built, this `Partial`
  instead results in a partially bound function or class:

      partial_config = Partial(TestClass)
      partial_config.arg = 1
      partial_config.kwarg = 'kwarg'
      partial_class = build(partial_config)
      instance = partial_class(arg=2)  # Keyword arguments can be overridden.
      assert instance.arg == 2
      assert instance.kwarg == 'kwarg'

  A `Partial` can also be created from an existing `Config`, by passing it to
  the `Partial` constructor. This results in a shallow copy that is decoupled
  from the `Config` used to create it. In the example below, any further changes
  to `partial_config` are not reflected by `class_config` (and vice versa):

      partial_config = Partial(class_config)
      class_config.arg = 'new value'  # Further modification to `class_config`.
      partial_class = build(partial_config)
      instance = partial_class()
      assert instance.arg == 'arg'  # The instance config is still 'arg'.
  """

  # NOTE: We currently need to repeat this annotation for pytype.
  __fn_or_cls__: TypeOrCallableProducingT

  def __build__(self, *args, **kwargs):
    """Builds this `Partial` for the given `args` and `kwargs`.

    This method is called during `build` to get the output for this `Partial`.

    Args:
      *args: Positional arguments to partially bind to `self.__fn_or_cls__`.
      **kwargs: Keyword arguments to partially bind to `self.__fn_or_cls__`.

    Returns:
      A `functools.partial` or instance with `args` and `kwargs` bound to
      `self.__fn_or_cls__`.
    """
    return functools.partial(self.__fn_or_cls__, *args, **kwargs)

  def __call__(self, *args, **kwargs):
    """Converts this `Partial` into a `Config`, with optional overrides.

    Args:
      *args: Positional arguments to configure for `self.__fn_or_cls__`.
      **kwargs: Keyword arguments to configure for `self.__fn_or_cls__`.

    Returns:
      A `Config` instance copied from this `Partial`. All arguments are copied
      from this `Partial` instance, and arguments may additionally be
      (re)configured by passing `args` or `kwargs`.
    """
    return Config(self, *args, **kwargs)


def update_callable(buildable: Buildable,
                    new_callable: TypeOrCallableProducingT):
  """Updates `config` to build `new_callable` instead.

  When extending a base configuration, it can often be useful to swap one class
  for another. For example, an experiment may want to swap in a subclass that
  has augmented functionality.

  `update_callable` updates `config` in-place (preserving argument history).

  Args:
    buildable: A `Buildable` (e.g. a `fdl.Config`) to mutate.
    new_callable: The new callable `config` should call when built.

  Raises:
    TypeError: if `new_callable` has varargs, or if there are arguments set on
      `config` that are invalid to pass to `new_callable`.
  """
  # TODO: Consider adding a "drop_invalid_args: bool = False" argument.

  # Note: can't just call config.__init__(new_callable, **config.__arguments__)
  # to preserve history.
  #
  # Note: can't call `setattr` on all the args to validate them, because that
  # will result in duplicate history entries.

  original_args = buildable.__arguments__
  signature = inspect.signature(new_callable)
  if any(param.kind == param.VAR_POSITIONAL
         for param in signature.parameters.values()):
    raise NotImplementedError(
        'Variable positional arguments (aka `*args`) not supported.')
  has_var_keyword = any(param.kind == param.VAR_KEYWORD
                        for param in signature.parameters.values())
  if not has_var_keyword:
    invalid_args = [
        arg for arg in original_args.keys() if arg not in signature.parameters
    ]
    if invalid_args:
      raise TypeError(f'Cannot switch to {new_callable} (from '
                      f'{buildable.__fn_or_cls__}) because the Buildable would '
                      f'have invalid arguments {invalid_args}.')

  object.__setattr__(buildable, '__fn_or_cls__', new_callable)
  object.__setattr__(buildable, '__signature__', signature)
  object.__setattr__(buildable, '_has_var_keyword', has_var_keyword)
  buildable.__argument_history__['__fn_or_cls__'].append(
      history.entry('__fn_or_cls__', new_callable))


def assign(buildable: Buildable, **kwargs):
  """Assigns multiple arguments to cfg.

  Although this function does not enable a caller to do something they can't
  already do with other syntax, this helper function can be useful when
  manipulating deeply nested configs. Example:

  ```py
  cfg = # ...
  fdl.assign(cfg.my.deeply.nested.child.object, arg_a=1, arg_b='b')
  ```

  The above code snippet is equivalent to:

  ```py
  cfg = # ...
  cfg.my.deeply.nested.child.object.arg_a = 1
  cfg.my.deeply.nested.child.object.arg_b = 'b'
  ```

  Args:
    buildable: A `Buildable` (e.g. a `fdl.Config`) to set values upon.
    **kwargs: The arguments and values to assign.
  """
  for name, value in kwargs.items():
    setattr(buildable, name, value)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Types to make up some very important business logic.

These are types that will be configured via Fiddle's command line args.

Importantly: there is no dependency on Fiddle in this file!
"""

import dataclasses


@dataclasses.dataclass
class MyLinearModel:
  w: float
  b: float
  activation_dtype: str

  def __call__(self, x: float) -> float:
    return x * self.w + self.b


class MyDataLoader:

  def __init__(self, filename: str):
    pass

  def __iter__(self):
    return iter([1.0])  # Empty iterator


@dataclasses.dataclass
class BulkInferenceRunner:
  model: MyLinearModel
  data: MyDataLoader

  def run(self):
    print('Running...')
    for example in self.data:
      print(self.model(example))
    print('All done!')



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for selectors."""

import dataclasses
from typing import Any

from absl.testing import absltest
import fiddle as fdl
from fiddle.experimental import selectors


def fake_init_fn(rng, shape, dtype):
  del rng, shape, dtype  # unused
  return 1234


@dataclasses.dataclass
class Attention:
  dtype: Any
  kernel_init: Any
  bias_init: Any


@dataclasses.dataclass
class CrossAttention(Attention):
  """Example subclass."""


@dataclasses.dataclass
class Mlp:
  dtype: Any
  use_bias: bool


# In real models there are multiple layers, but pretend there is just one for
# this test.
@dataclasses.dataclass
class FakeEncoder:
  attention: Attention
  mlp: Mlp


@dataclasses.dataclass
class FakeDecoder:
  self_attention: Attention
  encoder_decoder_attention: Attention
  mlp: Mlp


@dataclasses.dataclass
class FakeEncoderDecoder:
  encoder: FakeEncoder
  decoder: FakeDecoder


def encoder_decoder_config() -> fdl.Config[FakeEncoderDecoder]:
  # This config node would usually not be shared, but is here so that we can
  # test how seen nodes are only visited once for efficiency.
  bias_init = fdl.Partial(fake_init_fn, None, None, None)
  encoder_cfg = fdl.Config(
      FakeEncoder,
      attention=fdl.Config(Attention, "float32", "kernel1", bias_init),
      mlp=fdl.Config(Mlp, "float32", False),
  )
  decoder_cfg = fdl.Config(
      FakeDecoder,
      self_attention=fdl.Config(Attention, "float32", "kernel1", bias_init),
      encoder_decoder_attention=fdl.Config(CrossAttention, "float32", "kernel2",
                                           "bias2"),
      mlp=fdl.Config(Mlp, "float32", False),
  )
  return fdl.Config(FakeEncoderDecoder, encoder_cfg, decoder_cfg)


class SelectionTest(absltest.TestCase):

  def test_matches_everything(self):
    cfg = encoder_decoder_config()
    sel = selectors.select(cfg)
    self.assertTrue(sel._matches(cfg.encoder))
    self.assertTrue(sel._matches(cfg.encoder.attention))
    self.assertTrue(sel._matches(cfg.encoder.mlp))
    self.assertTrue(sel._matches(cfg.decoder))
    self.assertTrue(sel._matches(cfg.decoder.self_attention))
    self.assertTrue(sel._matches(cfg.decoder.encoder_decoder_attention))
    self.assertTrue(sel._matches(cfg.decoder.mlp))

  def test_matches_based_on_type(self):
    cfg = encoder_decoder_config()
    sel = selectors.select(cfg, Attention)
    self.assertFalse(sel._matches(cfg.encoder))
    self.assertTrue(sel._matches(cfg.encoder.attention))
    self.assertFalse(sel._matches(cfg.encoder.mlp))

    # Matches a subclass.
    self.assertTrue(sel._matches(cfg.decoder.encoder_decoder_attention))

  def test_matches_function_call(self):
    cfg = encoder_decoder_config()
    for cfg_node in selectors.select(cfg, fake_init_fn):
      self.assertIs(cfg_node.__fn_or_cls__, fake_init_fn)

  def test___iter__(self):
    cfg = encoder_decoder_config()

    # There are 3 attention instances, 1 for the encoder and 2 for the decoder.
    self.assertLen(list(selectors.select(cfg, Attention)), 3)

    # There are 2 attention instances that are exactly the Attention class.
    self.assertLen(
        list(selectors.select(cfg, Attention, match_subclasses=False)), 2)

    # There is only one cross-attention instance.
    self.assertLen(list(selectors.select(cfg, CrossAttention)), 1)

    # The shared kernel init node is only visited once.
    self.assertLen(list(selectors.select(cfg, fake_init_fn)), 1)

  def test_setattr(self):
    cfg = encoder_decoder_config()
    selectors.select(cfg, Attention).dtype = "override_dtype"
    self.assertEqual(cfg.encoder.attention.dtype, "override_dtype")
    self.assertEqual(cfg.decoder.self_attention.dtype, "override_dtype")
    self.assertEqual(cfg.decoder.encoder_decoder_attention.dtype,
                     "override_dtype")
    self.assertEqual(cfg.encoder.mlp.dtype, "float32")

  def test_set(self):
    cfg = encoder_decoder_config()
    selectors.select(cfg, Attention).set(
        dtype="override_dtype", kernel_init="override_init")
    self.assertEqual(cfg.encoder.attention.dtype, "override_dtype")
    self.assertEqual(cfg.decoder.self_attention.dtype, "override_dtype")
    self.assertEqual(cfg.encoder.attention.kernel_init, "override_init")
    self.assertEqual(cfg.decoder.self_attention.kernel_init, "override_init")

  def test_debug_get(self):
    cfg = encoder_decoder_config()
    attention_kernels = list(
        selectors.select(cfg, Attention).get("kernel_init"))
    self.assertCountEqual(["kernel1", "kernel1", "kernel2"], attention_kernels)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for module_reflection."""

import types

from absl.testing import absltest
from fiddle import module_reflection
from fiddle import module_reflection_test_module as test_module
from fiddle.experimental import auto_config


class ModuleReflectionTest(absltest.TestCase):

  def test_find_fiddler_like_things(self):
    ns = types.SimpleNamespace()
    ns.simple = lambda x: None
    ns.not_one = lambda x, y: None
    ns.another = lambda y: None
    ns.zero_arity = lambda: None

    def kwarg_fn(cfg, **kwargs):  # pylint: disable=unused-argument
      pass

    ns.kwarg_fn = kwarg_fn

    def with_defaults(cfg, defaulted_arg=1):  # pylint: disable=unused-argument
      pass

    ns.with_defaults = with_defaults

    expected = sorted(['simple', 'another', 'kwarg_fn', 'with_defaults'])
    self.assertEqual(expected, module_reflection.find_fiddler_like_things(ns))

  def test_find_base_config_like_things(self):
    ns = types.SimpleNamespace()
    ns.base = lambda: None
    ns.another = lambda: None
    ns.fiddler = lambda x: None
    ns.not_me = lambda x, y: None

    expected = sorted(['base', 'another'])
    self.assertEqual(expected,
                     module_reflection.find_base_config_like_things(ns))

  def test_find_base_config_like_things_module(self):
    expected = sorted(['simple_base', 'alternate_base', 'base_with_defaults'])
    self.assertEqual(
        expected, module_reflection.find_base_config_like_things(test_module))

  def test_find_fiddler_like_things_module(self):
    expected = sorted(['fiddler1', 'fiddler2', 'another_fiddler'])
    self.assertEqual(expected,
                     module_reflection.find_fiddler_like_things(test_module))

  def test_auto_config_functions(self):

    def my_fn(x, y=3):
      return (x, y)

    @auto_config.auto_config
    def base_config():
      return my_fn(1, 3)

    ns = types.SimpleNamespace()
    ns.base_config = base_config

    expected = ['base_config']
    self.assertIn('base_config', dir(ns))
    self.assertEqual(expected,
                     module_reflection.find_base_config_like_things(ns))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for visualization of Fiddle configurations.

By default, you should be able to call APIs in `graphviz` and `printing` with no
special handling. But some real-world configurations become gigantic, and so
having a few helper functions for trimming them for interactive demos and such
can be valuable.
"""
import copy
from typing import Any, Dict, List

from fiddle import config
from fiddle import graphviz
import tree


def _raise_error():
  """Helper function for `Trimmed` that raises an error."""
  raise ValueError('Please do not call fdl.build() on a config tree with '
                   'Trimmed() nodes. These nodes are for visualization only.')


class Trimmed(config.Config[type(None)], graphviz.CustomGraphvizBuildable):

  def __init__(self):
    super().__init__(_raise_error)

  def __render_value__(self, api: graphviz.GraphvizRendererApi) -> Any:
    return api.tag('i')('(trimmed...)')


def trimmed(cfg: config.Buildable,
            trim: List[config.Buildable]) -> config.Buildable:
  """Returns a deep copy of a Buildable, with certain nodes replaced.

  This copy should have the same DAG structure as the original configuration.

  Example usage:

  graphviz.render(visualize.trimmed(cfg, [cfg.my_large_model_sub_config]))

  Args:
    cfg: Root buildable object.
    trim: List of sub-buildables that will be replaced by Trimmed() instances.

  Returns:
    Copy of the Buildable with nodes replaced.
  """

  return copy.deepcopy(cfg, {id(sub_cfg): Trimmed() for sub_cfg in trim})


def depth_over(cfg: config.Buildable, depth: int) -> List[config.Buildable]:
  """Returns a list of sub-Buildable objects over a given depth from the root.

  If a sub-Buildable has multiple paths from the root to it, its depth is the
  *minimum* distance of those paths.

  Args:
    cfg: Root buildable object.
    depth: Depth value.

  Returns:
    List of all sub-Buildables with depth strictly greater than `depth`.
  """
  node_to_depth: Dict[int, int] = {}  # Buildable hash --> min depth.
  id_to_node: Dict[int, config.Buildable] = {}  # Buildable hash --> Buildable.

  def _make_depth_map(node: config.Buildable, curr_depth: int):
    """Mapper function to generate the `node_to_depth` map."""
    if id(node) in node_to_depth:
      node_to_depth[id(node)] = min(node_to_depth[id(node)], curr_depth)
    else:
      node_to_depth[id(node)] = curr_depth
    id_to_node[id(node)] = node

    def leaf_fn(leaf):
      if isinstance(leaf, config.Buildable):
        _make_depth_map(leaf, curr_depth + 1)

    tree.map_structure(leaf_fn, node.__arguments__)

  _make_depth_map(cfg, 0)

  return [
      id_to_node[key]
      for key, node_depth in node_to_depth.items()
      if node_depth > depth
  ]



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a renderer to visualize a DAG of `fdl.Buildable`s via Graphviz."""

import abc
import html
import itertools

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from fiddle import config as fdl
from fiddle import tagging
from fiddle.codegen import formatting_utilities
import graphviz
import typing_extensions

_BUILDABLE_INSTANCE_COLORS = [
    '#ffc0cb',  # pink
    '#90ee90',  # lightgreen
    '#fff8dc',  # cornsilk
    '#ffa07a',  # lightsalmon
    '#add8e6',  # lightblue
    '#ff8c00',  # darkorange
    '#8fbc8f',  # darkseagreen
    '#adff2f',  # greenyellow
    '#ff6347',  # tomato
    '#db7093',  # palevioletred
    '#f0e68c',  # khaki
    '#32cd32',  # limegreen
    '#00bfff',  # deepskyblue
    '#7b68ee',  # mediumslateblue
]


class GraphvizRendererApi(typing_extensions.Protocol):
  """API of _GraphvizRenderer exposed to CustomGraphvizBuildable subclasses."""

  def tag(self, tag: str, **kwargs) -> Callable[[Any], str]:
    raise NotImplementedError()


class CustomGraphvizBuildable(metaclass=abc.ABCMeta):
  """Mixin class that marks a Buildable has having a custom __render_value__.

  This lets certain special-purpose Buildables customize how they are rendered.
  """

  @abc.abstractmethod
  def __render_value__(self, api: GraphvizRendererApi) -> Any:
    """Renders this Buildable as a value."""


class _GraphvizRenderer:
  """Encapsulates state maintained while rendering a `Config` to Graphviz."""

  def __init__(self,
               instance_colors: Optional[List[str]] = None,
               max_sequence_elements_per_row: int = 10):
    """Initializes the render.

    Args:
      instance_colors: Optional list of HTML hexadecimal color codes (e.g.
        '#a1b2c3') to override the default set used to assign colors to
        `fdl.Buildable` instances encountered during rendering. Colors are
        assigned in order (and repeated if there are more instances than
        provided colors).
      max_sequence_elements_per_row: When rendering sequences, up to this many
        elements will be included in a single row of the output table, with
        additional rows added to render remaining elements.
    """
    self._instance_colors = instance_colors or _BUILDABLE_INSTANCE_COLORS
    self._max_sequence_elements_per_row = max_sequence_elements_per_row
    self._attr_defaults = {
        # Should contain a mapping of tag names to dicts of attribute defaults.
        'table': {
            'border': 0,
            'cellspacing': 0,
            'cellborder': 1,
            'cellpadding': 3
        },
    }
    self._clear()

  def _clear(self):
    """Resets all state associated with this renderer."""
    self._dot = graphviz.Graph(
        graph_attr={
            'overlap': 'false',
        },
        node_attr={
            'fontname': 'Courier',
            'fontsize': '10',
            'shape': 'none',
            'margin': '0',
        },
        edge_attr={
            'color': '#00000030',
            'penwidth': '3',
        })
    self._configs = {}
    # The id of the config currently being rendered. This is obtained from
    # _config_counter and used in _render_config and _render_nested_config.
    self._current_id = None
    self._config_counter = itertools.count()
    # Used to assign unique port names to table cells for nested configs.
    self._port_counter = itertools.count()

  def _color(self, index: int):
    return self._instance_colors[index % len(self._instance_colors)]

  def tag(self, tag: str, **kwargs) -> Callable[[Any], str]:
    """Returns a function that creates HTML tags of type `tag`.

    Example:

        td = self.tag('td', colspan=2)
        td('cell contents') => '<td colspan="2">cell contents</td>

    Args:
      tag: The type of tag to create (e.g. 'table', 'font', etc).
      **kwargs: Attributes to apply to the tag. Tag-specific default values for
        attributes can be provided in `self._attr_defaults`. The values provided
        here in **kwargs take precedence over defaults.

    Returns:
      A function that can be called with contents to put inside the tag, and
      returns HTML markup enclosing the contents inside `tag`. If the given
      contents is a list or tuple, the list elements are automatically
      concatenated. Otherwise, the contents are simply formatted as a string.
    """
    attr_values = {**self._attr_defaults.get(tag, {}), **kwargs}
    attributes = [
        f' {key.replace("_", "-")}="{value}"' if value is not None else ''
        for key, value in attr_values.items()
    ]

    def tag_fn(contents: Any) -> str:
      if isinstance(contents, (list, tuple)):
        contents = ''.join(contents)
      return f'<{tag}{"".join(attributes)}>{contents}</{tag}>'

    return tag_fn

  def _header_row(self, header, colspan=2, bgcolor='#eeeeee', style='solid'):
    """Constructs a header table row."""
    tr = self.tag('tr')
    header_td = self.tag('td', colspan=colspan, bgcolor=bgcolor, style=style)
    return tr(header_td(header))

  def render(self, config: fdl.Buildable) -> graphviz.Graph:
    """Renders the given `config`, recursively rendering any nested configs."""
    self._render_config(config)
    dot = self._dot
    self._clear()
    return dot

  def _render_config(self, config):
    """Adds `config` to the graph and renders it as a table."""
    if id(config) in self._configs:
      return  # Don't do anything if `config` has already been rendered.

    last_id = self._current_id
    self._current_id = next(self._config_counter)
    self._configs[id(config)] = self._current_id

    # Generate the header row.
    bgcolor = self._color(self._current_id)
    style = 'dashed' if isinstance(config, fdl.Partial) else 'solid'
    type_font = self.tag('font', point_size=8)
    type_name = config.__class__.__name__
    if isinstance(config, tagging.TaggedValue):
      key_names = ', '.join(repr(key.name) for key in config.tags)
      title = (type_font(html.escape(f'{type_name}: {key_names}')) + '&nbsp;')
      header = self._header_row(title, colspan=1, bgcolor=bgcolor, style=style)
    else:
      fn_or_cls_name = config.__fn_or_cls__.__name__
      title = (
          type_font(html.escape(f'{type_name}:')) + '&nbsp;' +
          html.escape(fn_or_cls_name))
      header = self._header_row(title, bgcolor=bgcolor, style=style)

    # Generate the arguments table.
    if isinstance(config, tagging.TaggedValue):
      table = self.tag('table')
      tr = self.tag('tr')
      value_td = self.tag('td', align='left')
      label = table([header, tr([value_td(self._render_value(config.value))])])
    elif config.__arguments__:
      label = self._render_dict(
          config.__arguments__, header=header, key_format_fn=str)
    else:
      table = self.tag('table')
      italics = self.tag('i')
      label = table([header, self._header_row(italics('no arguments'))])
    self._dot.node(str(self._current_id), f'<{label}>')

    self._current_id = last_id

  def _render_value(self, value: Any):
    """Renders an arbitrary value inside a `Config` rendering."""
    if value is tagging.NO_VALUE:
      return self.tag('i')('tagging.NO_VALUE')
    elif isinstance(value, CustomGraphvizBuildable):
      return value.__render_value__(self)
    elif isinstance(value, fdl.Buildable):
      return self._render_nested_config(value)
    elif isinstance(value, dict):
      return self._render_dict(
          value, header=self._header_row(type(value).__name__))
    elif isinstance(value, tuple) and hasattr(value, '_asdict'):
      # If it has an "_asdict" attribute it's almost certainly a `NamedTuple`.
      # Unfortunately there is no direct way of testing whether an object is a
      # `NamedTuple`, since all `NamedTuple`s are direct subclasses of `tuple`.
      return self._render_dict(
          value._asdict(),
          header=self._header_row(type(value).__name__),
          key_format_fn=str)
    elif isinstance(value, (list, tuple)):
      return self._render_sequence(value)
    else:
      return self._render_leaf(value)

  def _render_nested_config(self, config: fdl.Buildable) -> str:
    """Renders a nested config as a rectangle, linked to its parent config.

    This function renders the appearance of a config when it is nested inside
    another config (and potentially inside a list/dictionary within the config).
    First, `config` is rendered as its own separate unnested graph node (if it
    hasn't been already). The nested rendering is then just a single-celled
    table whose color matches the header color of the separate node rendering.
    The nested cell is then connected via an edge to the separate rendering.

    Args:
      config: The nested `fdl.Buildable` to render.

    Returns:
      The HTML markup for the resulting table representing `config`.
    """
    # First, add the config to the graph. This will add a separate node and
    # render it as a separate table (but is a no-op if it's already added).
    self._render_config(config)
    # Look up the config id for the config, and get a new unique port name to
    # use for the nested cell (see below).  The port allows the edge to go
    # "inside" the parent config and connect directly to the table cell.
    config_id = self._configs[id(config)]
    port = next(self._port_counter)
    # Now add an edge in the graph to the parent config. The direction here
    # determines the order in which the graph is layed out when using the
    # default "dot" layout engine, so putting the parent config first lays the
    # graph out from root to children.
    self._dot.edge(f'{self._current_id}:{port}:c', f'{config_id}:c')

    # Return a table with a single colored cell, using the port name from above.
    style = 'dashed' if isinstance(config, fdl.Partial) else 'solid'
    table = self.tag('table', style=style)
    tr = self.tag('tr')
    td = self.tag('td', port=port, bgcolor=self._color(config_id), style=style)
    return table(tr(td('')))

  def _render_sequence(self, sequence: Union[List[Any], Tuple[Any]]) -> str:
    """Renders the given sequence (list or tuple) as an HTML table."""
    table = self.tag('table')
    tr = self.tag('tr')
    td = self.tag('td')
    index_td = self.tag('td', cellpadding=0, bgcolor='#eeeeee')
    index_font = self.tag('font', point_size=6)

    if not sequence:
      return '[]' if isinstance(sequence, list) else '()'

    cells, indices = [], []
    for i, value in enumerate(sequence):
      cells.append(td(self._render_value(value)))
      indices.append(index_td(index_font(i)))
    row_stride = self._max_sequence_elements_per_row
    colspan = min(len(sequence), row_stride)
    rows = [self._header_row(type(sequence).__name__, colspan=colspan)]
    for i in range(0, len(sequence), row_stride):
      rows.extend([tr(cells[i:i + row_stride]), tr(indices[i:i + row_stride])])
    return table(rows)

  def _render_dict(self,
                   dict_: Dict[str, Any],
                   header: Optional[str] = None,
                   key_format_fn: Callable[[Any], str] = repr) -> str:
    """Renders the given `dict_` as an HTML table.

    Args:
      dict_: The `dict` to render.
      header: A table row containing a header row to include. The table row's
        table cell should have colspan="2" to render properly in the table.
      key_format_fn: A function to apply to dictionary keys to conver them into
        a string representation. Defaults to `repr`.

    Returns:
      The HTML markup for the resulting table representing `dict_`.
    """
    table = self.tag('table')
    tr = self.tag('tr')
    key_td = self.tag('td', align='right', bgcolor='#eeeeee')
    value_td = self.tag('td', align='left')

    rows = [header] if header is not None else []
    for key, value in dict_.items():
      key = key_format_fn(key)
      rows.append(tr([key_td(key), value_td(self._render_value(value))]))
    return table(rows)

  def _render_leaf(self, value: Any) -> str:
    """Renders `value` as its `__repr__` string."""
    value = formatting_utilities.pretty_print(value)
    return html.escape(repr(value))


def render(config: fdl.Buildable) -> graphviz.Graph:
  """Renders the given `config` as a `graphviz.Graph`.

  Each config is rendered as a table of keys and values (with a header row).
  Any nested configs get their own separate table, with an edge pointing to
  their location in their parent config. If a config instance is present in
  multiple parent configs, it is only rendered once but will have multiple
  edges to parent configs.

  Args:
    config: The config to render.

  Returns:
    A `graphviz.Graph` object containing the resulting rendering of `config`.
    Standard `graphviz` methods can then be used to export this to a file.
  """
  return _GraphvizRenderer().render(config)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tags attach metadata to arguments & values for ergonomic fiddling.

When defining shared parameters across a project that later could be changed,
for example dtype or activation function, we encourage the following coding
pattern: a tiny shared library file should declare tags for an
entire project, like

  class ActivationDType(fdl.Tag):
    "Data type (e.g. jnp.float32, or jnp.bfloat16) for intermediate values."
  class KernelInitFn(fdl.Tag):
    "A function from RNG & hyperparameters to initial weights."

Then, in library code which configures shared Fiddle fixtures, these tags are
used,

  def layer_norm_fixture() -> fdl.Config[LayerNorm]:
    cfg = fdl.Config(LayerNorm)
    cfg.dtype = ActivationDType.new(default=jnp.float32)

And in experiment code stitching everything together, all of these tagged values
can be set at once,

  def encoder_decoder_fixture() -> fdl.Config[EncoderDecoder]:
    cfg = fdl.Config(EncoderDecoder)
    ...
    cfg.encoder.encoder_norm = layer_norm_fixture()

    # Set all activation dtypes.
    fdl.set_tagged(cfg, ActivationDType, jnp.bfloat16)
    return cfg

  model = fdl.build(encoder_decoder_fixture())

Tags can be defined in a module (file) or inside classes, but cannot be defined
within a function, as there would not be a way to refer to it without
re-invoking the function, defining a new unique Tag.

Tags can inherit from each other. Use this power wisely.

While the normal mechanism for creating a TaggedValue is via $TAG.new(), you can
also create a TaggedValue explicitly, allowing multiple tags to be associated
with the value.
"""

from __future__ import annotations

import copy
import inspect
from typing import Any, Collection, FrozenSet, Generic, Set, TypeVar, Union

from fiddle import config
import tree


class TaggedValueNotFilledError(ValueError):
  """A TaggedValue was not filled when build() was called."""


class _NoValue:
  """Sentinel class (used in place of object for more precise errors)."""

  def __deepcopy__(self, memo):
    """Override for deepcopy that does not copy this sentinel object."""
    del memo
    return self


NO_VALUE = _NoValue()


class TagType(type):
  """All Fiddle tags are instances of this class.

  For defining Tags, we leverage Python's class definition and documentation
  syntax.

  See the documentation on `fdl.Tag` for instructions on how to use.
  """

  def __init__(cls, name, bases, dct):
    if '__doc__' not in dct:
      raise TypeError('You must provide a tag description with a docstring.')
    if '__qualname__' not in dct:
      raise TypeError('No `__qualname__` property found.')
    if '<' in dct['__qualname__']:
      raise TypeError('You cannot define a tag within a function or lambda.')
    super().__init__(name, bases, dct)

  def __call__(cls, *args, **kwds):
    raise TypeError('You cannot instantiate Fiddle tags (trying to instantiate '
                    f'{cls.name}); just use the type itself (no parenthesis) '
                    'when specifying tags on a TaggedValue(...) call, or call '
                    '`.new(default=)` to create a new `TaggedValue`.')

  @property
  def description(cls) -> str:
    """A string describing the semantics and intended usecases for this tag."""
    return cls.__doc__

  @property
  def name(cls) -> str:
    """A unique name for this tag."""
    return f'{cls.__module__}.{cls.__qualname__}'

  def __str__(cls) -> str:
    return f'#{cls.name}'

  def __repr__(cls) -> str:  # pylint: disable=invalid-repr-returned
    return cls.name


class Tag(metaclass=TagType):
  """Metadata associated with a Fiddle configurable value.

  To see a usage example, please see the documentation on the tagging module.

  Note: Tags cannot be instantiated.
  """

  @classmethod
  def new(cls, default: Any = NO_VALUE) -> TaggedValue:
    """Creates a new `TaggedValue` with `cls` as the only tag.

    If you would like to create a `TaggedValue` with multiple tags attached,
    directly call the `TaggedValue` constructor:

    ```py
    tagged_value = TaggedValue(tags=(Tag1, Tag2, Tag3), default=Foo)
    ```

    Args:
      default: Optional. The default value for this TaggedValue.

    Returns:
      A TaggedValue tagged with the tag `cls`.
    """
    return TaggedValue(tags=(cls,), default=default)


def tagvalue_fn(tags: Set[TagType], value: Any = NO_VALUE) -> Any:
  if value is NO_VALUE:
    raise TaggedValueNotFilledError(
        'Expected all `TaggedValue`s to be replaced via fdl.set_tagged() '
        f'calls, but one with tags {tags} was not set.')
  else:
    return value


T = TypeVar('T')


class TaggedValue(Generic[T], config.Config[T]):
  """Declares a value annotated with a set of `Tag`s."""

  def __init__(
      self,
      tags: Collection[TagType],
      default: Union[_NoValue, T] = NO_VALUE,
  ):
    """Initializes the TaggedValue.

    Args:
      tags: A non-empty collection of tags to associated with this value.
      default: Default value of the TaggedValue. By default this is a sentinel
        which will cause the configuration to fail to build when the
        placeholders are not set.
    """
    if isinstance(tags, TaggedValue):
      # Handle copy.copy call (`type(self)(self)`).
      other = tags
      super().__init__(tagvalue_fn, tags=other.tags, value=other.value)
      return
    if not tags:
      raise ValueError('At least one tag must be provided.')
    super().__init__(tagvalue_fn, tags=set(tags), value=default)

  def __deepcopy__(self, memo) -> config.Buildable[T]:
    """Implements the deepcopy API."""
    return TaggedValue(tags=self.tags, default=copy.deepcopy(self.value, memo))


# TODO: Migrate users of this API to a `select`-based API.
def set_tagged(root: config.Buildable, *, tag: TagType, value: Any) -> None:
  """Sets all parameters in `root` tagged with `tag` to `value`.

  Args:
    root: The root of a DAG of Buildables.
    tag: The tag to search for.
    value: Value to set for all parameters tagged with `tag`.
  """
  if isinstance(root, TaggedValue):
    if any(issubclass(t, tag) for t in root.tags):
      root.value = value

  def map_fn(leaf):
    if isinstance(leaf, config.Buildable):
      set_tagged(root=leaf, tag=tag, value=value)
    return leaf

  # TODO: Use something other than map to avoid creating unused garbage.
  tree.map_structure(map_fn, root.__arguments__)


def list_tags(
    root: config.Buildable,
    add_superclasses=False,
) -> FrozenSet[TagType]:
  """Lists all tags in a buildable.

  Args:
    root: The root of a DAG of `Buildable`s.
    add_superclasses: For tags that inherit from other tags, add the
      superclasses as well.

  Returns:
    Set of tags used in this buildable.
  """
  tags = set()

  def _inner(node: config.Buildable):
    if isinstance(node, TaggedValue):
      tags.update(node.tags)

    def map_fn(leaf):
      if isinstance(leaf, config.Buildable):
        _inner(leaf)

    # TODO: Use something other than map for efficiency.
    tree.map_structure(map_fn, node.__arguments__)

  _inner(root)

  # Add superclasses if desired.
  if add_superclasses:
    for tag in list(tags):
      for base in inspect.getmro(tag):
        if base not in tags and base is not Tag and issubclass(base, Tag):
          tags.add(base)

  return frozenset(tags)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for materialize."""

from absl.testing import absltest
from fiddle import config
from fiddle import materialize


class MaterializeTest(absltest.TestCase):

  def test_materialize_defaults(self):

    def test_defaulting_fn(x: int, y=3, z: str = 'abc'):
      del x, y, z  # Unused.

    cfg = config.Config(test_defaulting_fn)

    self.assertEqual({}, cfg.__arguments__)
    materialize.materialize_defaults(cfg)
    self.assertEqual({'y': 3, 'z': 'abc'}, cfg.__arguments__)

  def test_materialize_defaults_recursive(self):

    class Inner:

      def __init__(self, a, b: str = 'b'):
        self.a = a
        self.b = b

    def test_defaulting_fn(x, y, z: str = 'abc'):
      del x, y, z  # Unused.

    cfg = config.Config(test_defaulting_fn)
    cfg.y = [
        config.Config(Inner, 'a'),
        config.Config(Inner),
    ]
    self.assertEqual({'a': 'a'}, cfg.y[0].__arguments__)
    self.assertEqual({}, cfg.y[1].__arguments__)

    materialize.materialize_defaults(cfg, recurse=False)
    self.assertEqual({'a': 'a'}, cfg.y[0].__arguments__)
    self.assertEqual({}, cfg.y[1].__arguments__)
    self.assertEqual('abc', cfg.__arguments__['z'])

    materialize.materialize_defaults(cfg, recurse=True)
    self.assertEqual({'a': 'a', 'b': 'b'}, cfg.y[0].__arguments__)
    self.assertEqual({'b': 'b'}, cfg.y[1].__arguments__)
    self.assertEqual('abc', cfg.__arguments__['z'])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for serialization."""

import pickle

from absl.testing import absltest
import fiddle as fdl
from fiddle.experimental import serialization


class Unserializable:

  def __getstate__(self):
    raise NotImplementedError()


def identity_fn(arg1):
  return arg1


class SerializationTest(absltest.TestCase):

  def test_pickling_non_serializable_history_deepcopy_default(self):
    cfg = fdl.Config(identity_fn, arg1=Unserializable())
    cfg.arg1 = 4
    serialization.clear_argument_history(cfg)
    with self.assertRaises(NotImplementedError):
      pickle.dumps(cfg)

  def test_pickling_non_serializable_history(self):
    cfg = fdl.Config(identity_fn, arg1=Unserializable())
    cfg.arg1 = 4
    cfg = serialization.clear_argument_history(cfg)
    pickle.dumps(cfg)

  def test_pickling_non_serializable_history_mutation(self):
    cfg = fdl.Config(identity_fn, arg1=Unserializable())
    cfg.arg1 = 4
    serialization.clear_argument_history(cfg, deepcopy=False)
    pickle.dumps(cfg)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements Fiddle's build() function."""
import contextlib
import threading

from typing import Any, Dict, List, Tuple

from fiddle import config
from fiddle import tagging


import tree


class _BuildGuardState(threading.local):

  def __init__(self):
    super().__init__()
    self.in_build = False


_state = _BuildGuardState()


@contextlib.contextmanager
def _in_build():
  """A context manager to ensure fdl.build is not called recursively."""
  if _state.in_build:
    raise ValueError(
        'It is forbidden to call `fdl.build` inside another `fdl.build` call.')
  _state.in_build = True
  try:
    yield
  finally:
    _state.in_build = False


class BuildError(ValueError):
  """Error raised when building a Config fails."""

  def __init__(
      self,
      buildable: config.Buildable,
      path_from_config_root: str,
      original_error: Exception,
      args: Tuple[Any, ...],
      kwargs: Dict[str, Any],
  ) -> None:
    super().__init__(str(original_error))
    self.buildable = buildable
    self.path_from_config_root = path_from_config_root
    self.original_error = original_error
    self.args = args
    self.kwargs = kwargs

  def __str__(self):
    fn_or_cls_name = self.buildable.__fn_or_cls__.__qualname__
    return (f'Failed to construct or call {fn_or_cls_name} '
            f'(at {self.path_from_config_root}) with arguments\n'
            f'    args: {self.args}\n'
            f'    kwargs: {self.kwargs}')


# This is a free function instead of a method on the `Buildable` object in order
# to avoid potential naming collisions (e.g., if a function or class has a
# parameter named `build`).
def build(buildable):
  """Builds `buildable`, recursively building any nested `Buildable` instances.

  This is the core function for turning a `Buildable` into a usable object. It
  recursively walks through `buildable`'s parameters, building any nested
  `Config` instances. Depending on the specific `Buildable` type passed
  (`Config` or `Partial`), the result is either the result of calling
  `config.__fn_or_cls__` with the configured parameters, or a partial function
  or class with those parameters bound.

  If the same `Buildable` instance is seen multiple times during traversal of
  the configuration tree, `build` is called only once (for the first instance
  encountered), and the result is reused for subsequent copies of the instance.
  This is achieved via the `memo` dictionary (similar to `deepcopy`). This has
  the effect that for configured class instances, each separate config instance
  is in one-to-one correspondence with an actual instance of the configured
  class after calling `build` (shared config instances <=> shared class
  instances).

  Args:
    buildable: A `Buildable` instance to build.

  Returns:
    The built version of `buildable`.
  """
  memo = {}

  def _build(buildable, path_str: str):

    def map_fn(map_path: List[Any], leaf):
      attr, *rest = map_path
      leaf_path = f'{path_str}.{attr}' + ''.join(f'[{x!r}]' for x in rest)
      is_buildable = isinstance(leaf, config.Buildable)
      return _build(leaf, leaf_path) if is_buildable else leaf

    if id(buildable) not in memo:
      kwargs = tree.map_structure_with_path(map_fn, buildable.__arguments__)
      try:
        memo[id(buildable)] = buildable.__build__(**kwargs)
      except tagging.TaggedValueNotFilledError:
        raise
      except Exception as e:
        raise BuildError(buildable, path_str, e, (), kwargs) from e

    return memo[id(buildable)]

  with _in_build():
    return _build(buildable, '<root>')



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fiddle extensions to handle SeqIO code more elegantly.

(See https://github.com/google/seqio.)

Currently this just affects codegen, graphviz, and other debugging functions.
"""

from fiddle.codegen import codegen
from fiddle.codegen import mini_ast

_seqio_import = mini_ast.DirectImport(name="seqio")

_import_aliases = (
    # Seqio  aliases all of these to its top level import.
    ("seqio.dataset_providers", _seqio_import),
    ("seqio.evaluation", _seqio_import),
    ("seqio.feature_converters", _seqio_import),
    ("seqio.loggers", _seqio_import),
    ("seqio.utils", _seqio_import),
    ("seqio.vocabularies", _seqio_import),
)


def enable():
  """Registers SeqIO fiddle extensions.

  This allows for nicer handling of seqio imports.
  """
  for module_str, import_stmt in _import_aliases:
    codegen.register_import_alias(module_str, import_stmt)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the `fiddle.config` module."""

import copy
import functools
import pickle
import threading
from typing import Any, Callable, Dict

from absl.testing import absltest
from fiddle import building
from fiddle import config
from fiddle import history

import pytype_extensions

import tree


class TestClass:
  arg1: Any
  arg2: Any
  kwarg1: Any
  kwarg2: Any

  def __init__(self, arg1, arg2, kwarg1=None, kwarg2=None):  # pylint: disable=unused-argument
    self.__dict__.update(locals())

  def a_method(self):
    return 4  # A random number (https://xkcd.com/221/)


def basic_fn(arg1, arg2, kwarg1=None, kwarg2=None) -> Dict[str, Any]:  # pylint: disable=unused-argument
  return locals()


def fn_with_var_args(arg1, *args, kwarg1=None):  # pylint: disable=unused-argument
  return locals()


def fn_with_var_kwargs(arg1, kwarg1=None, **kwargs):  # pylint: disable=unused-argument
  return locals()


def fn_with_var_args_and_kwargs(arg1, *args, kwarg1=None, **kwargs):  # pylint: disable=unused-argument
  return locals()


class FiddleInitStaticMethod:

  def __init__(self, x, y=3, **kwargs):
    self.x = x
    self.y = y
    for key, value in kwargs.items():
      setattr(self, key, value)

  @staticmethod
  def __fiddle_init__(cfg):
    cfg.z = 5


class FiddleInitStaticMethodChild(FiddleInitStaticMethod):

  @classmethod
  def __fiddle_init__(cls, cfg):
    super().__fiddle_init__(cfg)
    cfg.w = 42


class FiddleInitClassMethod:

  def __init__(self, x, y=3, **kwargs):
    self.x = x
    self.y = y
    for key, value in kwargs.items():
      setattr(self, key, value)

  @classmethod
  def __fiddle_init__(cls, cfg):
    cfg.z = 5


class FiddleInitClassMethodChild(FiddleInitClassMethod):

  @classmethod
  def __fiddle_init__(cls, cfg):
    super().__fiddle_init__(cfg)
    cfg.w = 10


class FiddleInitIncompatibleChild(FiddleInitClassMethod):

  def __init__(self, a, b):
    self.a = a
    self.b = b
    super().__init__(x=a + 1, y=b - 3, z=a + b)

  @classmethod
  def __fiddle_init__(cls, cfg):
    # Since FiddleInitClassMethod has different init arguments, we cannot call
    # super().__fiddle_init__(cfg).
    super_cfg = config.Config(FiddleInitClassMethod)
    cfg.b = super_cfg.y + 3  # Demonstrate copying over.


class FiddleInitIncompatibleChildBroken(FiddleInitClassMethod):

  def __init__(self, a, b):  # pylint: disable=super-init-not-called
    pass


def make_typed_config() -> config.Config[TestClass]:
  """Helper function which returns a config.Config whose type is known."""
  return config.Config(TestClass, arg1=1, arg2=2)


def make_untyped_config(arg_to_configure, **kwargs) -> config.Config:
  """Helper function which returns an untyped config.Config."""
  return config.Config(arg_to_configure, **kwargs)


def make_typed_partial() -> config.Partial[TestClass]:
  """Helper function to create a typed Partial instance."""
  return config.Partial(TestClass, arg1=1)


def make_untyped_partial(arg_to_configure, **kwargs) -> config.Partial:
  """Helper function which returns an untyped config.Config."""
  return config.Partial(arg_to_configure, **kwargs)


class Unserializable:

  def __getstate__(self):
    raise NotImplementedError()


def _test_fn_unserializable_default(x=Unserializable()):
  return x


class ConfigTest(absltest.TestCase):

  def test_config_for_classes(self):
    class_config = config.Config(TestClass, 1, kwarg2='kwarg2')
    pytype_extensions.assert_type(class_config, config.Config[TestClass])
    self.assertEqual(class_config.arg1, 1)
    self.assertEqual(class_config.kwarg2, 'kwarg2')
    class_config.arg1 = 'arg1'
    self.assertEqual(class_config.arg1, 'arg1')
    class_config.arg2 = 'arg2'
    class_config.kwarg1 = 'kwarg1'

    instance = building.build(class_config)
    pytype_extensions.assert_type(instance, TestClass)
    self.assertEqual(instance.arg1, 'arg1')
    self.assertEqual(instance.arg2, 'arg2')
    self.assertEqual(instance.kwarg1, 'kwarg1')
    self.assertEqual(instance.kwarg2, 'kwarg2')

  def test_config_for_functions(self):
    fn_config = config.Config(basic_fn, 1, kwarg2='kwarg2')
    pytype_extensions.assert_type(fn_config, config.Config[Dict[str, Any]])
    self.assertEqual(fn_config.arg1, 1)
    self.assertEqual(fn_config.kwarg2, 'kwarg2')
    fn_config.arg1 = 'arg1'
    self.assertEqual(fn_config.arg1, 'arg1')
    fn_config.arg2 = 'arg2'
    fn_config.kwarg1 = 'kwarg1'

    result = building.build(fn_config)
    pytype_extensions.assert_type(result, Dict[str, Any])
    self.assertEqual(result, {
        'arg1': 'arg1',
        'arg2': 'arg2',
        'kwarg1': 'kwarg1',
        'kwarg2': 'kwarg2'
    })

  def test_config_for_functions_with_var_args(self):
    fn_config = config.Config(fn_with_var_args, 'arg1', kwarg1='kwarg1')
    pytype_extensions.assert_type(fn_config, config.Config)
    fn_args = building.build(fn_config)
    self.assertEqual(fn_args['arg1'], 'arg1')
    self.assertEqual(fn_args['kwarg1'], 'kwarg1')

    fn_config.arg1 = 'new_arg1'
    fn_config.kwarg1 = 'new_kwarg1'
    fn_args = building.build(fn_config)
    pytype_extensions.assert_type(fn_args, Any)
    self.assertEqual(fn_args['arg1'], 'new_arg1')
    self.assertEqual(fn_args['kwarg1'], 'new_kwarg1')

  def test_config_for_functions_with_var_kwargs(self):
    fn_config = config.Config(
        fn_with_var_kwargs,
        'arg1',
        kwarg1='kwarg1',
        kwarg2='kwarg2',
        kwarg3='kwarg3')
    fn_args = building.build(fn_config)
    self.assertEqual(fn_args['arg1'], 'arg1')
    self.assertEqual(fn_args['kwarg1'], 'kwarg1')
    self.assertEqual(fn_args['kwargs'], {
        'kwarg2': 'kwarg2',
        'kwarg3': 'kwarg3'
    })

    fn_config.kwarg1 = 'new_kwarg1'
    fn_config.kwarg3 = 'new_kwarg3'
    fn_args = building.build(fn_config)
    self.assertEqual(fn_args['kwarg1'], 'new_kwarg1')
    self.assertEqual(fn_args['kwargs'], {
        'kwarg2': 'kwarg2',
        'kwarg3': 'new_kwarg3'
    })

  def test_config_for_functions_with_var_args_and_kwargs(self):
    fn_config = config.Config(fn_with_var_args_and_kwargs, arg1='arg1')
    fn_args = building.build(fn_config)
    self.assertEqual(fn_args['arg1'], 'arg1')

    fn_config.args = 'kwarg_called_arg'
    fn_config.kwargs = 'kwarg_called_kwarg'
    fn_args = building.build(fn_config)
    self.assertEqual(fn_args['kwargs'], {
        'args': 'kwarg_called_arg',
        'kwargs': 'kwarg_called_kwarg'
    })

  def test_fiddle_init_config_static(self):
    cfg = config.Config(FiddleInitStaticMethod)
    self.assertEqual(5, cfg.z)
    cfg.x = 1
    obj = building.build(cfg)
    self.assertIsInstance(obj, FiddleInitStaticMethod)
    self.assertEqual(1, obj.x)
    self.assertEqual(3, obj.y)
    self.assertEqual(5, obj.z)  # pytype: disable=attribute-error

  def test_fiddle_init_partial_static(self):
    cfg = config.Partial(FiddleInitStaticMethod)
    self.assertEqual(5, cfg.z)
    partial = building.build(cfg)
    obj = partial(x=1)
    self.assertIsInstance(obj, FiddleInitStaticMethod)
    self.assertEqual(1, obj.x)
    self.assertEqual(3, obj.y)
    self.assertEqual(5, obj.z)  # pytype: disable=attribute-error

  def test_fiddle_init_config_static_subclass(self):
    cfg = config.Config(FiddleInitStaticMethodChild)
    self.assertEqual(3, cfg.y)
    self.assertEqual(5, cfg.z)
    self.assertEqual(42, cfg.w)
    cfg.x = 1
    obj = building.build(cfg)
    self.assertIsInstance(obj, FiddleInitStaticMethodChild)
    self.assertEqual(42, obj.w)  # pytype: disable=attribute-error
    self.assertEqual(1, obj.x)
    self.assertEqual(3, obj.y)
    self.assertEqual(5, cfg.z)  # pytype: disable=attribute-error

  def test_fiddle_init_config_class(self):
    cfg = config.Config(FiddleInitClassMethod, 1)
    self.assertEqual(5, cfg.z)
    obj = building.build(cfg)
    self.assertIsInstance(obj, FiddleInitClassMethod)
    self.assertEqual(1, obj.x)
    self.assertEqual(3, obj.y)
    self.assertEqual(5, obj.z)  # pytype: disable=attribute-error

  def test_fiddle_init_partial_class(self):
    cfg = config.Partial(FiddleInitClassMethod, 1)
    self.assertEqual(5, cfg.z)
    partial = building.build(cfg)
    obj = partial(x=1)
    self.assertIsInstance(obj, FiddleInitClassMethod)
    self.assertEqual(1, obj.x)
    self.assertEqual(3, obj.y)
    self.assertEqual(5, obj.z)  # pytype: disable=attribute-error

  def test_fiddle_init_config_subclass(self):
    cfg = config.Config(FiddleInitClassMethodChild, 0)
    self.assertEqual(5, cfg.z)
    self.assertEqual(10, cfg.w)
    obj = building.build(cfg)
    self.assertIsInstance(obj, FiddleInitClassMethodChild)
    self.assertEqual(0, obj.x)
    self.assertEqual(3, obj.y)
    self.assertEqual(5, obj.z)  # pytype: disable=attribute-error
    self.assertEqual(10, obj.w)  # pytype: disable=attribute-error

  def test_fiddle_init_incompatible_subclass(self):
    cfg = config.Config(FiddleInitIncompatibleChild)
    self.assertEqual(6, cfg.b)
    cfg.a = 8
    obj = building.build(cfg)
    self.assertIsInstance(obj, FiddleInitIncompatibleChild)
    self.assertEqual(8, obj.a)
    self.assertEqual(9, obj.x)
    self.assertEqual(6, obj.b)
    self.assertEqual(3, obj.y)
    self.assertEqual(14, obj.z)  # pytype: disable=attribute-error

  def test_fiddle_init_incompatible_broken(self):
    with self.assertRaisesRegex(
        TypeError, 'No parameter named.*; valid parameter names: a, b'):
      _ = config.Config(FiddleInitIncompatibleChildBroken)

  def test_config_with_default_args(self):

    def my_func(x: int = 2, y: str = 'abc'):  # pylint: disable=unused-argument
      return locals()

    cfg = config.Config(my_func)
    self.assertEqual(2, cfg.x)
    self.assertEqual('abc', cfg.y)

    obj = building.build(cfg)
    self.assertEqual(2, obj['x'])
    self.assertEqual('abc', obj['y'])

    cfg.y = 'xyz'
    obj = building.build(cfg)
    self.assertEqual(2, obj['x'])
    self.assertEqual('xyz', obj['y'])

  def test_config_defaults_not_materialized(self):

    def my_func(x: int, y: str = 'abc', z: float = 2.0):  # pylint: disable=unused-argument
      return locals()

    cfg = config.Config(my_func)

    self.assertEqual('abc', cfg.y)  # Should return the default.
    self.assertEqual({}, cfg.__arguments__)  # but not materialized.

    cfg.x = 42
    output = building.build(cfg)

    self.assertEqual({'x': 42, 'y': 'abc', 'z': 2.0}, output)

  def test_nested_configs(self):
    fn_config1_args = {
        'arg1': 'innermost1',
        'arg2': 'innermost2',
        'kwarg1': 'kw1',
        'kwarg2': 'kw2'
    }
    fn_config1 = config.Config(basic_fn, *fn_config1_args.values())

    class_config = config.Config(
        TestClass, arg1=config.Partial(fn_config1), arg2=fn_config1)
    fn_config2 = config.Config(
        basic_fn, arg1=config.Partial(class_config), arg2=class_config)

    fn_config2_args = building.build(fn_config2)

    test_class_partial = fn_config2_args['arg1']
    test_class_instance = test_class_partial()
    self.assertEqual(type(test_class_instance), TestClass)
    self.assertEqual(test_class_instance.arg1(), fn_config1_args)
    self.assertEqual(test_class_instance.arg2, fn_config1_args)

    test_class_instance = fn_config2_args['arg2']
    self.assertEqual(type(test_class_instance), TestClass)
    self.assertEqual(test_class_instance.arg1(), fn_config1_args)
    self.assertEqual(test_class_instance.arg2, fn_config1_args)

  def test_instance_sharing(self):
    class_config = config.Config(
        TestClass, 'arg1', 'arg2', kwarg1='kwarg1', kwarg2='kwarg2')
    class_config_copy = copy.copy(class_config)
    class_config_copy.arg1 = 'separate_arg1'

    # Changing copied config parameters doesn't change the original config.
    self.assertEqual(class_config.arg1, 'arg1')

    fn_config = config.Config(basic_fn, class_config_copy, {
        'key1': [class_config, class_config],
        'key2': (class_config,)
    })

    fn_args = building.build(fn_config)
    separate_instance = fn_args['arg1']
    shared_instance = fn_args['arg2']['key1'][0]
    structure = fn_args['arg2']

    self.assertIsNot(shared_instance, separate_instance)
    for leaf in tree.flatten(structure):
      self.assertIs(leaf, shared_instance)

    self.assertEqual(shared_instance.arg1, 'arg1')
    self.assertEqual(separate_instance.arg1, 'separate_arg1')

  @absltest.expectedFailure  # TODO: Fix instance sharing bug.
  def test_instance_sharing_collections(self):
    child_configs = [
        config.Config(basic_fn, 1, 'a'),
        config.Config(basic_fn, 2, 'b')
    ]
    cfg = config.Config(TestClass)
    cfg.arg1 = child_configs
    cfg.arg2 = child_configs
    obj = building.build(cfg)

    self.assertIsInstance(obj, TestClass)
    self.assertIs(obj.arg1, obj.arg2)
    self.assertIs(obj.arg1[0], obj.arg2[0])
    self.assertIs(obj.arg1[1], obj.arg2[1])

  def test_shallow_copy(self):
    class_config = config.Config(TestClass, 'arg1', 'arg2')
    fn_config = config.Config(basic_fn, class_config, 'fn_arg2')
    fn_config_copy = copy.copy(fn_config)
    # Changing the copy doesn't change the original.
    fn_config_copy.arg2 = 'fn_arg2_copy'
    self.assertEqual(fn_config.arg2, 'fn_arg2')
    # But modifying a shared value is seen by both.
    fn_config_copy.arg1.arg2 = 'mutated'
    self.assertEqual(fn_config.arg1.arg2, 'mutated')

  def test_deep_copy(self):
    class_config = config.Config(TestClass, 'arg1', 'arg2')
    fn_config = config.Config(basic_fn, class_config, 'fn_arg2')
    fn_config_copy = copy.deepcopy(fn_config)
    # Changing the copy doesn't change the original.
    fn_config_copy.arg2 = 'fn_arg2_copy'
    self.assertEqual(fn_config.arg2, 'fn_arg2')
    # With a deep copy, the value is no longer shared.
    fn_config_copy.arg1.arg2 = 'mutated'
    self.assertEqual(fn_config.arg1.arg2, 'arg2')

  def test_deep_copy_preserves_instance_sharing(self):
    class_config = config.Config(TestClass, 'arg1', 'arg2')
    fn_config = config.Config(basic_fn, arg1=class_config, arg2=class_config)
    self.assertIs(fn_config.arg1, fn_config.arg2)
    fn_config_copy = copy.deepcopy(fn_config)
    self.assertIsNot(fn_config.arg1, fn_config_copy.arg1)
    self.assertIs(fn_config_copy.arg1, fn_config_copy.arg2)

  def test_deep_copy_partials(self):
    class_partial = config.Partial(TestClass, 'arg1', 'arg2')
    fn_config = config.Config(
        basic_fn, arg1=class_partial(), arg2=class_partial())
    self.assertIsNot(fn_config.arg1, fn_config.arg2)
    fn_config_copy = copy.deepcopy(fn_config)
    self.assertIsNot(fn_config_copy.arg1, fn_config_copy.arg2)
    self.assertIsNot(fn_config.arg1, fn_config_copy.arg1)

  def test_equality(self):
    cfg1 = config.Config(TestClass, 'arg1')
    cfg2 = config.Config(TestClass, 'arg1')
    self.assertEqual(cfg1, cfg2)
    cfg1.arg1 = 'arg2'
    self.assertNotEqual(cfg1, cfg2)

    partial = config.Partial(TestClass, 'arg1')
    self.assertNotEqual(cfg2, partial)

  def test_unsetting_argument(self):
    fn_config = config.Config(basic_fn)
    fn_config.arg1 = 3
    fn_config.arg2 = 4

    del fn_config.arg1

    with self.assertRaisesRegex(AttributeError,
                                "No parameter 'arg1' has been set"):
      _ = fn_config.arg1

    with self.assertRaisesRegex(AttributeError, "'arg1'"):
      del fn_config.arg1

    with self.assertRaisesRegex(AttributeError, "'unknown_arg'"):
      del fn_config.unknown_arg

    self.assertEqual(4, fn_config.arg2)

  def test_dir_simple(self):
    fn_config = config.Config(basic_fn)
    self.assertEqual(['arg1', 'arg2', 'kwarg1', 'kwarg2'], dir(fn_config))

  def test_dir_cls(self):
    cfg = config.Config(TestClass)
    self.assertEqual(['arg1', 'arg2', 'kwarg1', 'kwarg2'], dir(cfg))

  def test_dir_var_args_and_kwargs(self):
    varargs_config = config.Config(fn_with_var_args_and_kwargs)
    varargs_config.abc = '123'
    self.assertEqual(['abc', 'arg1', 'kwarg1'], dir(varargs_config))

  def test_partial_for_classes(self):
    class_partial = config.Partial(TestClass, 'arg1', 'arg2')
    pytype_extensions.assert_type(class_partial, config.Partial[TestClass])
    partial = building.build(class_partial)
    pytype_extensions.assert_type(partial, Callable[..., TestClass])
    self.assertIsInstance(partial, functools.partial)

    instance = partial()
    self.assertIsInstance(instance, TestClass)
    self.assertEqual(instance.arg1, 'arg1')
    self.assertEqual(instance.arg2, 'arg2')

    # We can override parameters at the call site.
    instance = partial(arg1='new_arg1', arg2='new_arg2', kwarg1='new_kwarg1')
    self.assertEqual(instance.arg1, 'new_arg1')
    self.assertEqual(instance.arg2, 'new_arg2')
    self.assertEqual(instance.kwarg1, 'new_kwarg1')

  def test_partial_for_functions(self):
    fn_partial = config.Partial(basic_fn, 1, kwarg2='kwarg2')
    pytype_extensions.assert_type(fn_partial, config.Partial[Dict[str, Any]])
    self.assertEqual(fn_partial.arg1, 1)
    self.assertEqual(fn_partial.kwarg2, 'kwarg2')
    fn_partial.arg1 = 'arg1'
    self.assertEqual(fn_partial.arg1, 'arg1')
    fn_partial.arg2 = 'arg2'
    fn_partial.kwarg1 = 'kwarg1'

    partial = building.build(fn_partial)
    pytype_extensions.assert_type(partial, Callable[..., Dict[str, Any]])
    self.assertEqual(partial(), {
        'arg1': 'arg1',
        'arg2': 'arg2',
        'kwarg1': 'kwarg1',
        'kwarg2': 'kwarg2'
    })

  def test_typed_config(self):
    class_config = make_typed_config()
    pytype_extensions.assert_type(class_config, config.Config[TestClass])
    instance = building.build(class_config)
    pytype_extensions.assert_type(instance, TestClass)
    self.assertEqual(instance.arg1, 1)
    self.assertEqual(instance.arg2, 2)

  def test_untyped_config(self):
    class_config = make_untyped_config(TestClass, arg1=2, arg2=3)
    pytype_extensions.assert_type(class_config, config.Config)
    instance = building.build(class_config)
    pytype_extensions.assert_type(instance, Any)
    self.assertEqual(instance.arg1, 2)
    self.assertEqual(instance.arg2, 3)

  def test_typed_partial(self):
    class_partial = make_typed_partial()
    pytype_extensions.assert_type(class_partial, config.Partial[TestClass])
    partial = building.build(class_partial)
    pytype_extensions.assert_type(partial, Callable[..., TestClass])
    instance = partial(arg2=4)
    self.assertEqual(instance.arg1, 1)
    self.assertEqual(instance.arg2, 4)

  def test_untyped_partial(self):
    class_partial = make_untyped_partial(TestClass, arg1=2)
    pytype_extensions.assert_type(class_partial, config.Partial)
    partial = building.build(class_partial)
    pytype_extensions.assert_type(partial, Callable[..., Any])
    instance = partial(arg2=4)
    self.assertEqual(instance.arg1, 2)
    self.assertEqual(instance.arg2, 4)

  def test_call_partial(self):
    class_partial = config.Partial(TestClass, 'arg1', 'arg2')
    class_config = class_partial('new_arg1', kwarg1='new_kwarg1')
    class_config.arg2 = 'new_arg2'
    self.assertEqual(class_partial.arg2, 'arg2')
    instance = building.build(class_config)
    self.assertEqual(instance.arg1, 'new_arg1')
    self.assertEqual(instance.arg2, 'new_arg2')
    self.assertEqual(instance.kwarg1, 'new_kwarg1')

  def test_call_partial_nested(self):
    class_partial = config.Partial(TestClass, 'arg1', 'arg2')
    class_config = config.Config(TestClass, class_partial(), class_partial())
    instance = building.build(class_config)
    self.assertEqual(instance.arg1.arg1, 'arg1')
    self.assertEqual(instance.arg1.arg2, 'arg2')
    self.assertEqual(instance.arg2.arg1, 'arg1')
    self.assertEqual(instance.arg2.arg2, 'arg2')
    self.assertIsNot(instance.arg1, instance.arg2)

  def test_repr_class_config(self):
    class_config = config.Config(TestClass, 1, 2, kwarg1='kwarg1')
    expected_repr = "<Config[TestClass(arg1=1, arg2=2, kwarg1='kwarg1')]>"
    self.assertEqual(repr(class_config), expected_repr)

  def test_repr_fn_config(self):
    fn_config = config.Config(basic_fn, 1, 2, kwarg1='kwarg1')
    expected_repr = "<Config[basic_fn(arg1=1, arg2=2, kwarg1='kwarg1')]>"
    self.assertEqual(repr(fn_config), expected_repr)

  def test_repr_class_partial(self):
    class_partial = config.Partial(TestClass, 1, 2, kwarg1='kwarg1')
    expected_repr = "<Partial[TestClass(arg1=1, arg2=2, kwarg1='kwarg1')]>"
    self.assertEqual(repr(class_partial), expected_repr)

  def test_repr_fn_partial(self):
    fn_partial = config.Partial(basic_fn, 1, 2, kwarg1='kwarg1')
    expected_repr = "<Partial[basic_fn(arg1=1, arg2=2, kwarg1='kwarg1')]>"
    self.assertEqual(repr(fn_partial), expected_repr)

  def test_nonexistent_attribute_error(self):
    class_config = config.Config(TestClass, 1)
    expected_msg = (r"No parameter 'nonexistent_arg' has been set on "
                    r'<Config\[TestClass\(arg1=1\)\]>\.')
    with self.assertRaisesRegex(AttributeError, expected_msg):
      getattr(class_config, 'nonexistent_arg')

  def test_nonexistent_parameter_error(self):
    class_config = config.Config(TestClass)
    expected_msg = (r"No parameter named 'nonexistent_arg' exists for "
                    r"<class '.*\.TestClass'>; valid parameter names: "
                    r'arg1, arg2, kwarg1, kwarg2\.')
    with self.assertRaisesRegex(TypeError, expected_msg):
      class_config.nonexistent_arg = 'error!'

  def test_nonexistent_var_args_parameter_error(self):
    fn_config = config.Config(fn_with_var_args)
    expected_msg = (r'Variadic arguments \(e.g. \*args\) are not supported\.')
    with self.assertRaisesRegex(TypeError, expected_msg):
      fn_config.args = (1, 2, 3)

  def test_unsupported_var_args_error(self):
    expected_msg = (r'Variable positional arguments \(aka `\*args`\) not '
                    r'supported\.')
    with self.assertRaisesRegex(NotImplementedError, expected_msg):
      config.Config(fn_with_var_args, 1, 2, 3)

  def test_build_inside_build(self):

    def inner_build(x: int) -> str:
      return str(x)

    def nest_call(x: int) -> str:
      cfg = config.Config(inner_build, x)
      return building.build(cfg)

    outer = config.Config(nest_call, 3)
    expected_msg = r'test_build_inside_build.<locals>.nest_call \(at <root>\)'
    with self.assertRaisesRegex(building.BuildError, expected_msg) as e:
      building.build(outer)
    expected_msg = r'forbidden to call `fdl.build` inside another `fdl.build`'
    self.assertRegex(str(e.exception.__cause__), expected_msg)

  def test_history_tracking(self):
    cfg = config.Config(TestClass, 'arg1_value')
    cfg.arg2 = 'arg2_value'
    del cfg.arg1

    self.assertEqual(
        set(['arg1', 'arg2', '__fn_or_cls__']),
        set(cfg.__argument_history__.keys()))
    self.assertLen(cfg.__argument_history__['arg1'], 2)
    self.assertLen(cfg.__argument_history__['arg2'], 1)
    self.assertEqual('arg1', cfg.__argument_history__['arg1'][0].param_name)
    self.assertEqual('arg1_value', cfg.__argument_history__['arg1'][0].value)
    self.assertRegex(
        str(cfg.__argument_history__['arg1'][0].location),
        r'config_test.py:\d+:test_history_tracking')
    self.assertEqual('arg2', cfg.__argument_history__['arg2'][0].param_name)
    self.assertEqual('arg2_value', cfg.__argument_history__['arg2'][0].value)
    self.assertRegex(
        str(cfg.__argument_history__['arg2'][0].location),
        r'config_test.py:\d+:test_history_tracking')
    self.assertEqual('arg1', cfg.__argument_history__['arg1'][1].param_name)
    self.assertEqual(history.DELETED, cfg.__argument_history__['arg1'][1].value)
    self.assertRegex(
        str(cfg.__argument_history__['arg1'][1].location),
        r'config_test.py:\d+:test_history_tracking')

    self.assertEqual(cfg.__argument_history__['arg1'][0].sequence_id + 1,
                     cfg.__argument_history__['arg2'][0].sequence_id)
    self.assertEqual(cfg.__argument_history__['arg2'][0].sequence_id + 1,
                     cfg.__argument_history__['arg1'][1].sequence_id)

  def test_custom_location_history_tracking(self):
    with history.custom_location(lambda: 'abc:123'):
      cfg = config.Config(TestClass, 'arg1')
    cfg.arg2 = 'arg2'
    self.assertEqual(
        set(['arg1', 'arg2', '__fn_or_cls__']),
        set(cfg.__argument_history__.keys()))
    self.assertLen(cfg.__argument_history__['arg1'], 1)
    self.assertLen(cfg.__argument_history__['arg2'], 1)
    self.assertEqual('arg1', cfg.__argument_history__['arg1'][0].param_name)
    self.assertRegex(
        str(cfg.__argument_history__['arg1'][0].location), 'abc:123')
    self.assertEqual('arg2', cfg.__argument_history__['arg2'][0].param_name)
    self.assertRegex(
        str(cfg.__argument_history__['arg2'][0].location),
        r'config_test.py:\d+:test_custom_location_history_tracking')
    self.assertEqual(cfg.__argument_history__['arg1'][0].sequence_id + 1,
                     cfg.__argument_history__['arg2'][0].sequence_id)

  def test_accessing_functions_on_config(self):
    """Test helpful error messages when users hold them wrong!

    Config objects should provide a helpful error message when users attempt
    to use them as if they were the actual built (underlying) objects.
    """

    cfg = config.Config(TestClass)
    expected_msg = 'a_method.*Note: .*TestClass has an attribute/method with '
    with self.assertRaisesRegex(AttributeError, expected_msg):
      cfg.a_method()
    with self.assertRaisesRegex(AttributeError, expected_msg):
      _ = cfg.a_method

  def test_unhashable(self):
    """All Buildable's should be unhashable: mutability and custom __eq__."""
    with self.assertRaisesRegex(TypeError, 'unhashable'):
      _ = config.Config(TestClass) in {}
    with self.assertRaisesRegex(TypeError, 'unhashable'):
      _ = config.Partial(TestClass) in {}

  def test_pickling_config(self):
    """Bulidable types should be pickle-able."""
    cfg = config.Config(TestClass, 1, 'abc')
    self.assertEqual(cfg, pickle.loads(pickle.dumps(cfg)))
    reloaded = pickle.loads(pickle.dumps(cfg))
    reloaded.kwarg1 = 3  # mutate after unpickling.
    self.assertNotEqual(cfg, reloaded)

  def test_pickling_partial(self):
    cfg = config.Partial(TestClass)
    cfg.arg1 = 'something'
    self.assertEqual(cfg, pickle.loads(pickle.dumps(cfg)))

  def test_pickling_composition(self):
    cfg = config.Config(TestClass, 1, 'abc')
    cfg.kwarg2 = config.Partial(TestClass)
    cfg.kwarg2.arg1 = 'something'
    self.assertEqual(cfg, pickle.loads(pickle.dumps(cfg)))

  def test_pickling_non_serializable_default(self):
    pickle.dumps(config.Config(_test_fn_unserializable_default))

  def test_build_raises_nice_error_too_few_args(self):
    cfg = config.Config(basic_fn, config.Config(TestClass, 1), 2)
    with self.assertRaisesRegex(building.BuildError, '.*TestClass.*') as e:
      building.build(cfg)
    self.assertIs(e.exception.buildable, cfg.arg1)
    self.assertEqual(e.exception.path_from_config_root, '<root>.arg1')
    self.assertIsInstance(e.exception.original_error, TypeError)
    self.assertEqual(e.exception.args, ())
    self.assertEqual(e.exception.kwargs, {'arg1': 1})

  def test_build_raises_exception_on_call(self):

    def raise_error():
      raise ValueError('My fancy exception')

    cfg = config.Config(raise_error)
    with self.assertRaisesWithLiteralMatch(
        building.BuildError, 'Failed to construct or call ConfigTest.'
        'test_build_raises_exception_on_call.<locals>.raise_error '
        '(at <root>) with arguments\n    args: ()\n    kwargs: {}'):
      building.build(cfg)

  def test_build_error_path(self):
    # This will raise an error, because it doesn't have one arg populated.
    sub_cfg = config.Config(basic_fn, 1)
    sub_dict = {'a': 0, 'b': 2, 'c': sub_cfg, 'd': 10}
    cfg = config.Config(fn_with_var_kwargs, [1, sub_dict])

    with self.assertRaises(building.BuildError) as e:
      building.build(cfg)

    self.assertEqual(e.exception.path_from_config_root, "<root>.arg1[1]['c']")

  def test_multithreaded_build(self):
    """Two threads can each invoke config.build without interfering."""
    output = None
    background_entered_build = threading.Event()
    foreground_entered_build = threading.Event()

    def other_thread():
      nonlocal output

      def blocking_function(x):
        background_entered_build.set()
        foreground_entered_build.wait()
        return x

      cfg = config.Config(blocking_function, 3)
      output = building.build(cfg)

    def blocking_function(x):
      foreground_entered_build.set()
      background_entered_build.wait()
      return x

    cfg = config.Config(blocking_function, 1)
    thread = threading.Thread(target=other_thread)
    thread.start()
    obj = building.build(cfg)
    thread.join()
    self.assertEqual(1, obj)
    self.assertEqual(3, output)

  def test_update_callable(self):
    cfg = config.Config(basic_fn, 1, 'xyz', kwarg1='abc')
    config.update_callable(cfg, TestClass)
    cfg.kwarg2 = '123'
    obj = building.build(cfg)
    self.assertIsInstance(obj, TestClass)
    self.assertEqual(1, obj.arg1)
    self.assertEqual('xyz', obj.arg2)
    self.assertEqual('abc', obj.kwarg1)
    self.assertEqual('123', obj.kwarg2)

  def test_update_callable_invalid_arg(self):
    cfg = config.Config(fn_with_var_kwargs, abc='123', xyz='321')
    with self.assertRaisesRegex(TypeError,
                                r"have invalid arguments \['abc', 'xyz'\]"):
      config.update_callable(cfg, TestClass)

  def test_update_callable_new_kwargs(self):
    cfg = config.Config(TestClass)
    cfg.arg1 = 1
    config.update_callable(cfg, fn_with_var_kwargs)
    cfg.abc = '123'  # A **kwargs value should now be allowed.
    self.assertEqual({
        'arg1': 1,
        'kwarg1': None,
        'kwargs': {
            'abc': '123'
        }
    }, building.build(cfg))

  def test_update_callable_varargs(self):
    cfg = config.Config(fn_with_var_kwargs, 1, 2)
    with self.assertRaisesRegex(NotImplementedError,
                                'Variable positional arguments'):
      config.update_callable(cfg, fn_with_var_args_and_kwargs)

  def test_assign(self):
    cfg = config.Config(fn_with_var_kwargs, 1, 2)
    config.assign(cfg, a='a', b='b')
    self.assertEqual({
        'arg1': 1,
        'kwarg1': 2,
        'kwargs': {
            'a': 'a',
            'b': 'b'
        }
    }, building.build(cfg))

  def test_assign_wrong_argument(self):
    cfg = config.Config(basic_fn)
    with self.assertRaisesRegex(TypeError, 'not_there'):
      config.assign(cfg, arg1=1, not_there=2)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The materialize module makes implicit defaults explicit.

Fiddle Buildable's allow accessing argument defaults, but they are treated
differently by the systems compared to values that have been explicitly set
(e.g. by a user's configuration). (This enables users and tooling to quickly see
what has been configured explicitly.)

For when a configuration is to be persisted for longer-term archival,
materializing the defaulted values can make the configuration archive somewhat
more hermetic.
"""

from typing import Any

from fiddle import config as buildable
import tree


def materialize_defaults(config: buildable.Buildable, *, recurse: bool = True):
  """Explicitly sets values for defaulted arguments.

  Buildable makes a distinction between a defaulted argument that hasn't been
  set, and an argument that has been explicitly set to a value (that may or may
  not be the same as the argument's default value).

  However, there are some instances where it can be valuable to explicitly
  materialize the defaulted values (e.g. when persisting a configuration in a
  text format, or for certain visualizations).

  Calling `materialize_defaults` on a Buildable is the equivalent of doing
  `cfg.MY_ARG = cfg.MY_ARG` for all arguments with default values that have not
  yet had a value explicitly set.

  Args:
    config: The Buildable to materialize defaults on; materialization occurs in
      place, and mutates `cfg`. If you would like `cfg` to be unmodified, make a
      copy first with `copy.deepcopy(cfg)`.
    recurse: Whether to recurse into the arguments of `cfg` and materialize
      their defaults too.
  """
  for arg in config.__signature__.parameters.values():
    if arg.default is not arg.empty and arg.name not in config.__arguments__:
      setattr(config, arg.name, arg.default)
  if recurse:
    tree.traverse(_materialize_defaults_traverse_function, config.__arguments__)


def _materialize_defaults_traverse_function(leaf: Any):
  """Callback used when traversing a tree in `materialize_config`."""
  if isinstance(leaf, buildable.Buildable):
    materialize_defaults(leaf, recurse=True)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A separate module used by tagging_test."""

import fiddle as fdl


class ParameterDType(fdl.Tag):
  """All trainable parameter's dtype.

  To control mixed precision, or quantization, all trainable parameter's dtypes
  are annotated with the `ParameterDType` tag.
  """


class LinearParamDType(ParameterDType):
  """All linear layer's DTypes."""


class ActivationDType(fdl.Tag):
  """Scalar type for arrays of module outputs."""


class MyModel:
  """Some domain-specific tags nested inside another type."""

  class EncoderDType(fdl.Tag):
    """Parameter dtypes in encoder layers."""

  class DecoderDType(fdl.Tag):
    """Parameter dtypes in decoder layers."""



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library for manipulating DAGs."""

import abc
import dataclasses
from typing import Any, List, Mapping, Tuple, Union


class PathElement(metaclass=abc.ABCMeta):
  """Element of a path."""

  @abc.abstractproperty
  def code(self) -> str:
    """Generates code for accessing this path."""
    raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class ListOrTupleItem(PathElement):
  """Index into a list or tuple."""
  base_obj: Union[List[Any], Tuple[Any, ...]]
  index: int

  @property
  def code(self) -> str:
    return f"[{self.index}]"


@dataclasses.dataclass(frozen=True)
class DictItem(PathElement):
  """An item from a dictionary with a given key."""
  base_obj: Mapping[Any, Any]
  key: Any

  @property
  def code(self) -> str:
    return f"[{self.key!r}]"


@dataclasses.dataclass(frozen=True)
class AttributeItem(PathElement):
  """An attribute of an object."""
  base_obj: Any
  name: str

  @property
  def code(self) -> str:
    return f".{self.name}"


class BuildableAttributeItem(AttributeItem):
  """An attribute of a Buildable."""


Path = List[PathElement]



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for fiddle.extensions.seqio."""

from typing import List

from absl.testing import absltest
import fiddle as fdl
from fiddle.codegen import codegen
import fiddle.extensions.seqio
import seqio


def tokens(code: str) -> List[str]:
  return code.strip().split()


class SeqioTest(absltest.TestCase):

  def test_codegen(self):
    fiddle.extensions.seqio.enable()
    cfg = fdl.Config(seqio.Evaluator)
    code = "\n".join(codegen.codegen_dot_syntax(cfg).lines())
    expected = """
import fiddle as fdl
import seqio


def build_config():
  root = fdl.Config(seqio.Evaluator)

  return root
    """
    self.assertEqual(tokens(code), tokens(expected))


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup script for fiddle.

See https://github.com/google/fiddle for documentation.
"""

from setuptools import find_packages
from setuptools import setup

_VERSION = '0.1.0'

long_description = """
# Fiddle

Fiddle is a Python-first configuration library particularly well suited to ML
applications. Fiddle enables deep configurability of parameters in a program,
while allowing configuration to be expressed in readable and maintainable
Python code.

**Authors**: Dan Holtmann-Rice, Brennan Saeta, Sergio Guadarrama
"""

setup(
    name='fiddle-config',
    version=_VERSION,
    include_package_data=True,
    packages=find_packages(exclude=['docs']),  # Required
    package_data={'testdata': ['testdata/*.fiddle']},
    install_requires=[
        'dm-tree',
        'typing-extensions',
    ],
    extras_require={
        'testing': [
            'absl-py',
            'pytest',
            'pytype',
            'seqio',
        ]
    },
    description='Fiddle: A Python-first configuration library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/google/fiddle',
    author='The Fiddle Team',
    author_email='noreply@google.com',
    classifiers=[
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',

        # Pick your license as you wish
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',

        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    project_urls={  # Optional
        'Documentation': 'https://github.com/google/fiddle/docs',
        'Bug Reports': 'https://github.com/google/fiddle/issues',
        'Source': 'https://github.com/google/fiddle',
    },
    license='Apache 2.0',
    keywords='fiddle python configuration machine learning'
)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A Config that builds a `types.SimpleNamespace` from any argument."""

import types
from typing import Any

from fiddle import config


def _kwargs_to_namespace(**kwargs: Any) -> types.SimpleNamespace:
  return types.SimpleNamespace(**kwargs)


class NamespaceConfig(config.Config):
  """A Config that builds a `types.SimpleNamespace` accepting all arg names."""

  def __init__(self, *args, **kwargs):
    super().__init__(_kwargs_to_namespace, *args, **kwargs)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for autofill."""

import dataclasses

from absl.testing import absltest
from fiddle import building
from fiddle import config
from fiddle.experimental import autofill


class Child:

  def __init__(self, x: str):
    self.x = x

  @staticmethod
  def __fiddle_init__(cfg):
    cfg.x = 42


@dataclasses.dataclass
class ChildDC:
  a: int
  b: str


@dataclasses.dataclass
class ParentDC:
  child1: Child
  child2: ChildDC


class AutoFillTest(absltest.TestCase):

  def test_nothing_to_fill(self):
    cfg = config.Config(Child)
    autofill.autofill(cfg)  # No changes expected.

    self.assertEqual(cfg, config.Config(Child))

  def test_simple_autofill(self):
    cfg = config.Config(ParentDC)

    self.assertFalse(hasattr(cfg, 'child1'))
    self.assertFalse(hasattr(cfg, 'child2'))

    autofill.autofill(cfg)

    self.assertIsInstance(cfg.child1, config.Config)
    self.assertEqual(cfg.child1.__fn_or_cls__, Child)
    self.assertIsInstance(cfg.child2, config.Config)
    self.assertEqual(cfg.child2.__fn_or_cls__, ChildDC)

  def test_recursive_autofill(self):

    class Grandparent:

      def __init__(self, parent: ParentDC):
        self.parent = parent

    cfg = config.Config(Grandparent)
    self.assertFalse(hasattr(cfg, 'parent'))

    autofill.autofill(cfg)

    self.assertIsInstance(cfg.parent, config.Config)
    self.assertEqual(cfg.parent.__fn_or_cls__, ParentDC)
    self.assertIsInstance(cfg.parent.child1, config.Config)
    self.assertEqual(cfg.parent.child1.__fn_or_cls__, Child)
    self.assertIsInstance(cfg.parent.child2, config.Config)
    self.assertEqual(cfg.parent.child2.__fn_or_cls__, ChildDC)

  def test_skip_if_no_annotation(self):

    class Parent:

      def __init__(self, child: ChildDC, other_thing):
        self.child = child
        self.other_thing = other_thing

    cfg = config.Config(Parent)

    self.assertFalse(hasattr(cfg, 'child'))
    self.assertFalse(hasattr(cfg, 'other_thing'))

    autofill.autofill(cfg)

    self.assertIsInstance(cfg.child, config.Config)
    self.assertFalse(hasattr(cfg, 'other_thing'))

  def test_skip_if_set(self):
    cfg = config.Config(ParentDC)
    cfg.child1 = config.Config(ChildDC)  # Switch type.
    autofill.autofill(cfg)

    self.assertIsInstance(cfg.child1, config.Config)
    self.assertEqual(cfg.child1.__fn_or_cls__, ChildDC)
    self.assertIsInstance(cfg.child2, config.Config)
    self.assertEqual(cfg.child2.__fn_or_cls__, ChildDC)

  def test_skip_if_default(self):

    @dataclasses.dataclass
    class Defaulted:
      child1: Child
      child2: ChildDC = ChildDC(a=-1, b='static')

    cfg = config.Config(Defaulted)
    autofill.autofill(cfg)

    self.assertIsInstance(cfg.child1, config.Config)
    self.assertIsInstance(cfg.child2, ChildDC)  # Not a config instance!

    cfg.child1.x = 42
    obj = building.build(cfg)

    self.assertEqual(-1, obj.child2.a)
    self.assertEqual('static', obj.child2.b)

  def test_handle_args_and_kwargs(self):

    def test_fn(a: int, b: Child, c: ParentDC, *args: ChildDC, **kwargs: Child):  # pylint: disable=unused-argument
      return locals()

    cfg = config.Config(test_fn)

    self.assertFalse(hasattr(cfg, 'a'))
    self.assertFalse(hasattr(cfg, 'b'))
    self.assertFalse(hasattr(cfg, 'c'))
    self.assertFalse(hasattr(cfg, 'args'))
    self.assertFalse(hasattr(cfg, 'kwargs'))

    autofill.autofill(cfg)

    self.assertFalse(hasattr(cfg, 'a'))
    self.assertTrue(hasattr(cfg, 'b'))
    self.assertTrue(hasattr(cfg, 'c'))
    self.assertFalse(hasattr(cfg, 'args'))
    self.assertFalse(hasattr(cfg, 'kwargs'))

    cfg.a = 42
    cfg.c.child2.a = 'a'
    cfg.c.child2.b = 'b'
    obj = building.build(cfg)

    self.assertEqual(42, obj['a'])
    self.assertIsInstance(obj['c'], ParentDC)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fiddle extensions to handle JAX code more elegantly.

Currently this just affects codegen, graphviz, and other debugging functions.
"""

from fiddle.codegen import codegen
from fiddle.codegen import mini_ast
from fiddle.codegen import special_value_codegen
import jax
from jax import numpy as jnp


def _make_jnp_importable(name: str) -> special_value_codegen.Importable:
  return special_value_codegen.SingleImportable(
      "jax.numpy", lambda jnp_name: f"{jnp_name}.{name}")


_jnp_type_importables = (
    (jnp.bool_, _make_jnp_importable("bool_")),
    (jnp.uint8, _make_jnp_importable("uint8")),
    (jnp.uint16, _make_jnp_importable("uint16")),
    (jnp.uint32, _make_jnp_importable("uint32")),
    (jnp.uint64, _make_jnp_importable("uint64")),
    (jnp.int8, _make_jnp_importable("int8")),
    (jnp.int16, _make_jnp_importable("int16")),
    (jnp.int32, _make_jnp_importable("int32")),
    (jnp.int64, _make_jnp_importable("int64")),
    (jnp.bfloat16, _make_jnp_importable("bfloat16")),
    (jnp.float16, _make_jnp_importable("float16")),
    (jnp.float32, _make_jnp_importable("float32")),
    (jnp.float64, _make_jnp_importable("float64")),
    (jnp.complex64, _make_jnp_importable("complex64")),
    (jnp.complex128, _make_jnp_importable("complex128")),
)

_import_aliases = (
    # Rewrite internal import for JAX initializers.
    ("jax._src.nn.initializers",
     mini_ast.FromImport(name="initializers", parent="jax.nn")),
    (
        "jax.numpy",
        mini_ast.FromImportAs(name="jnp", parent="jax", module="numpy"),
    ),
)


def _make_jax_nn_importable(name: str) -> special_value_codegen.Importable:
  return special_value_codegen.SingleImportable(
      "jax", lambda jax_mod_name: f"{jax_mod_name}.nn.{name}")


_nn_type_importables = (
    (jax.nn.relu, _make_jax_nn_importable("relu")),
    (jax.nn.gelu, _make_jax_nn_importable("gelu")),
    (jax.nn.relu6, _make_jax_nn_importable("relu6")),
    (jax.nn.silu, _make_jax_nn_importable("silu")),
    (jax.nn.soft_sign, _make_jax_nn_importable("soft_sign")),
    (jax.nn.sigmoid, _make_jax_nn_importable("sigmoid")),
    (jax.nn.selu, _make_jax_nn_importable("selu")),
    (jax.nn.log_sigmoid, _make_jax_nn_importable("log_sigmoid")),
    (jax.nn.hard_tanh, _make_jax_nn_importable("hard_tanh")),
    (jax.nn.hard_swish, _make_jax_nn_importable("hard_swish")),
    (jax.nn.hard_silu, _make_jax_nn_importable("hard_silu")),
    (jax.nn.tanh, _make_jax_nn_importable("tanh")),
    (jax.nn.swish, _make_jax_nn_importable("swish")),
)


def enable():
  """Registers JAX fiddle extensions.

  This allows for things like nicer handling of jax.numpy dtypes.
  """
  for value, importable in _jnp_type_importables:
    special_value_codegen.register_exact_value(value, importable)

  for value, importable in _nn_type_importables:
    special_value_codegen.register_exact_value(value, importable)

  for module_str, import_stmt in _import_aliases:
    codegen.register_import_alias(module_str, import_stmt)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of Fiddle-based configurations and fiddlers."""

import fiddle as fdl
from fiddle.absl_flags.example import business_logic
from fiddle.absl_flags.example import tags


def simple(
    filename="data.txt") -> fdl.Config[business_logic.BulkInferenceRunner]:
  """A simple base configuration-generating function."""
  model = fdl.Config(
      business_logic.MyLinearModel,
      w=-3.8,
      b=0.5,
      activation_dtype=tags.ActivationDtype.new(default="float32"))
  loader = fdl.Config(business_logic.MyDataLoader, filename=filename)
  return fdl.Config(business_logic.BulkInferenceRunner, model, loader)


def swap_weight_and_bias(cfg: fdl.Config[business_logic.BulkInferenceRunner]):
  """A sample fiddler."""
  tmp = cfg.model.w
  cfg.model.w = cfg.model.b
  cfg.model.b = tmp



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to help serialize configuration."""

import collections
import copy

import fiddle as fdl
import tree


def clear_argument_history(config: fdl.Buildable, deepcopy: bool = True):
  """Creates a copy of a config, clearing its history.

  This can be useful when a Config's history contains a non-serializable value.

  Args:
    config: Fiddle Buildable.
    deepcopy: Whether to deepcopy the configuration before clearing its history.

  Returns:
    Deepcopy of `config` with `__argument_history__` set to empty.
  """

  def _clear_history(node: fdl.Buildable):
    object.__setattr__(node, "__argument_history__",
                       collections.defaultdict(list))

    for leaf in tree.flatten(node.__arguments__):
      if isinstance(leaf, fdl.Buildable):
        _clear_history(leaf)

  if deepcopy:
    config = copy.deepcopy(config)
  _clear_history(config)
  return config



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for history."""

from absl.testing import absltest
from fiddle import history


class HistoryTest(absltest.TestCase):

  def test_location_formatting(self):
    location = history.Location(
        filename="my_file.py", line_number=123, function_name=None)
    self.assertEqual(str(location), "my_file.py:123")
    location = history.Location(
        filename="my_other_file.py",
        line_number=321,
        function_name="make_config")
    self.assertEqual(str(location), "my_other_file.py:321:make_config")

  def test_entry_simple(self):
    entry = history.entry("x", 1)
    self.assertEqual(entry.param_name, "x")
    self.assertEqual(entry.value, 1)

  def test_entry_deletion(self):
    entry = history.entry("y", history.DELETED)
    self.assertEqual(entry.param_name, "y")
    self.assertEqual(entry.value, history.DELETED)

  def test_location_provider(self):
    entry = history.entry("x", 123)
    self.assertRegex(entry.location.filename, "history_test.py")
    self.assertEqual(entry.location.function_name, "test_location_provider")

    self.assertEqual(entry.param_name, "x")
    self.assertEqual(entry.value, 123)

  def test_custom_location_provider(self):
    e1 = history.entry("x", 1)
    with history.custom_location(
        lambda: history.Location("other.py", 123, "foo")):
      e2 = history.entry("y", 2)
    e3 = history.entry("z", 3)

    self.assertRegex(e1.location.filename, "history_test.py")
    self.assertEqual(e2.location.filename, "other.py")
    self.assertRegex(e1.location.filename, "history_test.py")

    self.assertEqual(e2.location.line_number, 123)
    self.assertEqual(e1.location.line_number + 4, e3.location.line_number)

    self.assertEqual(e1.location.function_name, "test_custom_location_provider")
    self.assertEqual(e2.location.function_name, "foo")
    self.assertEqual(e3.location.function_name, "test_custom_location_provider")


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""An example demonstrating Fiddle and absl_flags.

Run this example with the following command:

```sh
python3 -m fiddle.absl_flags.example.example \
  --fdl_config=simple \
  --fiddler=swap_weight_and_bias \
  --fdl.model.b=0.73
  --fdl.data.filename='"other.txt"'  # Alt syntax: --fdl.data.filename=\"b.txt\"
```
"""

from typing import Sequence

from absl import app
import fiddle as fdl
from fiddle import absl_flags
from fiddle import printing
from fiddle.absl_flags.example import configs


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise ValueError(f'Unexpected CLI arguments: {argv!r}')
  cfg = absl_flags.create_buildable_from_flags(configs)
  print(printing.as_str_flattened(cfg))
  runner = fdl.build(cfg)
  runner.run()


if __name__ == '__main__':
  app.run(main, flags_parser=absl_flags.flags_parser)



# coding=utf-8
# Copyright 2022 The Fiddle-Config Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A test module used in `module_reflection_test`."""

import fiddle as fdl


# Note: normally, this function should be defined in a different module!
def function_to_configure(x: int, y: str):
  return x, y


def simple_base():
  return fdl.Config(function_to_configure, x=2)


def alternate_base():
  return fdl.Config(function_to_configure, x=0)


def base_with_defaults(x: int = 5, negative: bool = False):
  if negative:
    x = -x
  return fdl.Config(function_to_configure, x=x)


def arbitrary_other_function(x, y):  # pylint: disable=unused-argument
  pass


def fiddler1(cfg: fdl.Config):  # pylint: disable=unused-argument
  pass


def fiddler2(cfg: fdl.Config):  # pylint: disable=unused-argument
  pass


def another_fiddler(cfg: fdl.Config, defaulted_arg=3):  # pylint: disable=unused-argument
  pass
